"""Various Functions powered by Numba"""

from __future__ import annotations

from typing import Any

import numpy as np
from numba import literal_unroll, njit, prange, types
from numba.extending import overload
from pandas.api.types import (
    is_datetime64_dtype,
    is_numeric_dtype,
    is_timedelta64_dtype,
)

# https://numba.discourse.group/t/uint64-vs-int64-indexing-performance-difference/1500
# indexing with unsigned integers offers more performance


@njit(nogil=True)
def _numba_less_than_base(arr: np.ndarray, value: Any):
    """
    Get earliest position in `arr`
    where arr[i] <= `value`
    """
    # adapted from numba/np/array_math.py
    min_idx = 0
    max_idx = len(arr)
    while min_idx < max_idx:
        # to avoid overflow
        mid_idx = min_idx + ((max_idx - min_idx) >> 1)
        _mid_idx = np.uintp(mid_idx)
        if arr[_mid_idx] < value:
            min_idx = mid_idx + 1
        else:
            max_idx = mid_idx
    return min_idx


@njit(nogil=True)
def _numba_greater_than_base(arr: np.ndarray, value: Any):
    """
    Get latest position in `arr`
    where arr[i] >= `value`
    """
    # adapted from numba/np/array_math.py
    min_idx = 0
    max_idx = len(arr)
    while min_idx < max_idx:
        # to avoid overflow
        mid_idx = min_idx + ((max_idx - min_idx) >> 1)
        _mid_idx = np.uintp(mid_idx)
        if arr[_mid_idx] > value:
            max_idx = mid_idx
        else:
            min_idx = mid_idx + 1
    return min_idx


@njit(nogil=True)
def _numba_less_than(
    arr: np.ndarray,
    value: Any,
    strict: bool = False,
):
    """
    Get earliest position in `arr`
    where arr[i] </<= `value`
    """
    min_idx = _numba_less_than_base(arr=arr, value=value)
    if min_idx == arr.size:
        return -1
    if strict & (value == arr[min_idx]):
        min_idx = _numba_greater_than_base(arr=arr, value=value)
    if min_idx == arr.size:
        return -1
    return min_idx


@njit(nogil=True)
def _numba_greater_than(arr: np.ndarray, value: Any, strict: bool = False):
    """
    Get latest position in `arr`
    where arr[i] >/>= `value`
    """
    min_idx = _numba_greater_than_base(arr=arr, value=value)
    if not min_idx:
        return -1
    if strict & (value == arr[min_idx - 1]):
        min_idx = _numba_less_than_base(arr=arr, value=value)
    if not min_idx:
        return -1
    return min_idx


@njit(nogil=True)
def _le_lt_join_not_range_join(
    source_array: np.ndarray,
    search_array: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    strict: bool,
    booleans: np.ndarray,
):
    """
    Compute results for le/lt join, not in a range join
    """
    counts = 0
    new_slice_starts = np.empty(slice_starts.size, dtype=np.intp)
    for num in prange(booleans.size):
        l1 = source_array[num]
        slice_start = slice_starts[num]
        slice_end = slice_ends[num]
        r1 = search_array[slice_start:slice_end]
        position = _numba_less_than(arr=r1, value=l1, strict=strict)
        if position == -1:
            counts += 1
            new_slice_starts[num] = slice_end
            booleans[num] = False
        else:
            new_slice_starts[num] = slice_start + position
    return booleans, new_slice_starts, counts


@njit(nogil=True)
def _le_lt_join_range_join(
    source_array: np.ndarray,
    search_array: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    strict: bool,
    booleans: np.ndarray,
):
    """
    Compute results for le/lt join, in a range join
    """
    counts = 0
    new_slice_starts = np.empty(slice_starts.size, dtype=np.intp)
    for num in prange(booleans.size):
        l1 = source_array[num]
        slice_start = slice_starts[num]
        slice_end = slice_ends[num]
        if not booleans[num]:
            continue
        r1 = search_array[slice_start:slice_end]
        position = _numba_less_than(arr=r1, value=l1, strict=strict)
        if position == -1:
            counts += 1
            new_slice_starts[num] = slice_end
            booleans[num] = False
        else:
            new_slice_starts[num] = slice_start + position
    return booleans, new_slice_starts, counts


@njit(nogil=True)
def _ge_or_gt_equi_join(
    source_array: np.ndarray,
    search_array: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    strict: bool,
    booleans: np.ndarray,
):
    """Compute results for a greater than or equal join"""
    counts = 0
    new_slice_ends = np.empty(slice_ends.size, dtype=np.intp)
    for num in prange(booleans.size):
        l1 = source_array[num]
        slice_start = slice_starts[num]
        slice_end = slice_ends[num]
        r1 = search_array[slice_start:slice_end]
        position = _numba_greater_than(arr=r1, value=l1, strict=strict)
        if position == -1:
            counts += 1
            new_slice_ends[num] = slice_start
            booleans[num] = False
        else:
            new_slice_ends[num] = slice_start + position
    return booleans, new_slice_ends, counts


@njit(nogil=True)
def _generate_final_indices_monotonic_equi_join(
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    booleans: np.ndarray,
    left_index: np.ndarray,
    right_index: np.ndarray,
):
    """
    Create final indices if monotonic
    """
    begin = 0
    total = (slice_ends - slice_starts).sum()
    r_index = np.empty(total, dtype=np.intp)
    l_index = np.empty(total, dtype=np.intp)
    for _num in range(slice_ends.size):
        num = np.uintp(_num)
        if not booleans[num]:
            continue
        start = slice_starts[num]
        end = slice_ends[num]
        for n in range(start, end):
            n_ = np.uintp(n)
            r_index[begin] = right_index[n_]
            l_index[begin] = left_index[num]
            begin += 1
    return l_index, r_index


@njit(nogil=True)
def _eq_join_tuple_compare(
    booleans: np.ndarray,
    indices: np.ndarray,
    left_index: np.ndarray,
    start_indices: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    tupled: tuple,
):
    """Comparison when tupled is present in an equi join"""

    for _tuple in literal_unroll(tupled):
        left_arr = _tuple[0]
        right_arr = _tuple[1]
        op = _tuple[2]
        for n in range(left_index.size):
            _n = np.uintp(n)
            if not booleans[_n]:
                continue
            start = starts[_n]
            end = ends[_n]
            pos = 0
            counter = 0
            left_val = left_arr[_n]
            ind = start_indices[_n]
            for nn in range(start, end):
                _nn = np.uintp(nn)
                right_val = right_arr[_nn]
                boolean = _compare(left_val, right_val, op)
                _ind = np.uintp(ind + pos)
                # pos should always increment
                # no matter what happens
                # with the conditionals below
                pos += 1
                boolean = indices[_ind] & boolean
                indices[_ind] = boolean
                counter += np.intp(boolean)
            boolean_ = counter > 0
            booleans[_n] &= boolean_

    return booleans, indices


@njit(nogil=True)
def _get_row_count_tupled_equi_join(
    left_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    start_indices: np.ndarray,
    indices: np.ndarray,
    booleans: np.ndarray,
):
    """Get row count for an equi join with a tupled"""
    counts = np.zeros(left_index.size, dtype=np.intp)
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not booleans[_n]:
            continue
        start = starts[_n]
        end = ends[_n]
        pos = 0
        counter = 0
        ind = start_indices[_n]
        for _ in range(start, end):
            _ind = np.uintp(ind + pos)
            # pos should always increment
            # no matter what happens
            # with the condition below
            pos += 1
            if not indices[_ind]:
                continue
            counter += 1
        counts[_n] = counter
    return left_index, counts


@njit(nogil=True)
def _generate_indices_equi_join_tupled(
    indices: np.ndarray,
    left_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    start_indices: np.ndarray,
    right_index: np.ndarray,
    booleans: np.ndarray,
):
    """Build indices for equi join with tupled"""
    total = indices.sum()
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    indexer = 0
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not booleans[_n]:
            continue
        start = starts[_n]
        end = ends[_n]
        pos = 0
        counter = 0
        ind = start_indices[_n]
        l_index = left_index[_n]
        for nn in range(start, end):
            _ind = np.uintp(ind + pos)
            # pos should always increment
            # no matter what happens
            # with the condition below
            pos += 1
            if not indices[_ind]:
                continue
            _nn = np.uintp(nn)
            _indexer = np.uintp(indexer)
            left_indices[_indexer] = l_index
            right_indices[_indexer] = right_index[_nn]
            indexer += 1
            if indexer == total:
                counter = 1
                break
        if counter == 1:
            break
    return left_indices, right_indices


@njit(parallel=True, cache=True)
def _numba_equi_single_le_ge_join(
    left_index: np.ndarray,
    right_index: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    arr1: np.ndarray,
    arr2: np.ndarray,
    strict: bool,
    less_than: bool,
    row_count: bool,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Get indices for an equi join
    and a less than/greater than join
    """
    length = left_index.size
    booleans = np.ones(length, dtype=np.bool_)
    if less_than:
        booleans, slice_starts, counts = _le_lt_join_not_range_join(
            source_array=arr1,
            search_array=arr2,
            slice_starts=slice_starts,
            slice_ends=slice_ends,
            strict=strict,
            booleans=booleans,
        )
    else:
        booleans, slice_ends, counts = _ge_or_gt_equi_join(
            source_array=arr1,
            search_array=arr2,
            slice_starts=slice_starts,
            slice_ends=slice_ends,
            strict=strict,
            booleans=booleans,
        )
    if counts == length:
        return None, None
    if row_count:
        return left_index, slice_ends - slice_starts
    return _generate_final_indices_monotonic_equi_join(
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        booleans=booleans,
        left_index=left_index,
        right_index=right_index,
    )


@njit(parallel=True, cache=True)
def _numba_equi_single_le_ge_tupled_join(
    left_index: np.ndarray,
    right_index: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    arr1: np.ndarray,
    arr2: np.ndarray,
    strict: bool,
    less_than: bool,
    row_count: bool,
    tupled: tuple,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Get indices for an equi join
    and a less than/greater than join,
    and a tupled
    """
    length = left_index.size
    booleans = np.ones(length, dtype=np.bool_)
    if less_than:
        booleans, slice_starts, counts = _le_lt_join_not_range_join(
            source_array=arr1,
            search_array=arr2,
            slice_starts=slice_starts,
            slice_ends=slice_ends,
            strict=strict,
            booleans=booleans,
        )
    else:
        booleans, slice_ends, counts = _ge_or_gt_equi_join(
            source_array=arr1,
            search_array=arr2,
            slice_starts=slice_starts,
            slice_ends=slice_ends,
            strict=strict,
            booleans=booleans,
        )

    if counts == length:
        return None, None

    l_booleans = booleans[:]
    _indices = np.cumsum(slice_ends - slice_starts)
    start_indices = np.empty(left_index.size, dtype=np.intp)
    start_indices[0] = 0
    start_indices[1:] = _indices[:-1]
    _indices = _indices[-1]
    indices = np.ones(_indices, dtype=np.bool_)
    starts_slice = slice_starts[:]
    ends_slice = slice_ends[:]

    l_booleans, indices = _eq_join_tuple_compare(
        booleans=l_booleans,
        indices=indices,
        left_index=left_index,
        start_indices=start_indices,
        starts=starts_slice,
        ends=ends_slice,
        tupled=tupled,
    )

    if not np.any(l_booleans):
        return None, None
    if row_count:
        return _get_row_count_tupled_equi_join(
            left_index=left_index,
            starts=starts_slice,
            ends=ends_slice,
            start_indices=start_indices,
            indices=indices,
            booleans=l_booleans,
        )
    return _generate_indices_equi_join_tupled(
        indices=indices,
        left_index=left_index,
        starts=starts_slice,
        ends=ends_slice,
        start_indices=start_indices,
        right_index=right_index,
        booleans=l_booleans,
    )


@njit(parallel=True, cache=True)
def _numba_equi_join_range_join_monotonic(
    left_index: np.ndarray,
    right_index: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    ge_arr1: np.ndarray,
    ge_arr2: np.ndarray,
    ge_strict: bool,
    le_arr1: np.ndarray,
    le_arr2: np.ndarray,
    le_strict: bool,
    row_count: bool,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Get indices for an equi join
    and a range join
    """
    length = left_index.size
    booleans = np.ones(length, dtype=np.bool_)
    booleans, slice_ends, counts = _ge_or_gt_equi_join(
        source_array=ge_arr1,
        search_array=ge_arr2,
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        strict=ge_strict,
        booleans=booleans,
    )
    if counts == length:
        return None, None

    booleans, slice_starts, counts = _le_lt_join_not_range_join(
        source_array=le_arr1,
        search_array=le_arr2,
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        strict=le_strict,
        booleans=booleans,
    )
    if counts == length:
        return None, None

    if row_count:
        return left_index, slice_ends - slice_starts
    # no need to run a comparison
    # since all groups are monotonic increasing
    # simply create left and right indices
    return _generate_final_indices_monotonic_equi_join(
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        booleans=booleans,
        left_index=left_index,
        right_index=right_index,
    )


@njit(parallel=True, cache=True)
def _numba_equi_join_range_join_multiple_monotonic(
    left_index: np.ndarray,
    right_index: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    ge_arr1: np.ndarray,
    ge_arr2: np.ndarray,
    ge_strict: bool,
    le_arr1: np.ndarray,
    le_arr2: np.ndarray,
    le_strict: bool,
    row_count: bool,
    tupled: tuple,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Get indices for an equi join
    and a range join
    """
    length = left_index.size
    booleans = np.ones(length, dtype=np.bool_)
    booleans, slice_ends, counts = _ge_or_gt_equi_join(
        source_array=ge_arr1,
        search_array=ge_arr2,
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        strict=ge_strict,
        booleans=booleans,
    )
    if counts == length:
        return None, None

    booleans, slice_starts, counts = _le_lt_join_not_range_join(
        source_array=le_arr1,
        search_array=le_arr2,
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        strict=le_strict,
        booleans=booleans,
    )
    if counts == length:
        return None, None

    _indices = np.cumsum(slice_ends - slice_starts)
    start_indices = np.empty(left_index.size, dtype=np.intp)
    start_indices[0] = 0
    start_indices[1:] = _indices[:-1]
    _indices = _indices[-1]
    indices = np.ones(_indices, dtype=np.bool_)
    starts_slice = slice_starts[:]
    ends_slice = slice_ends[:]
    l_booleans = booleans[:]

    l_booleans, indices = _eq_join_tuple_compare(
        booleans=l_booleans,
        indices=indices,
        left_index=left_index,
        start_indices=start_indices,
        starts=starts_slice,
        ends=ends_slice,
        tupled=tupled,
    )

    if not np.any(l_booleans):
        return None, None

    if row_count:
        return _get_row_count_tupled_equi_join(
            left_index=left_index,
            starts=starts_slice,
            ends=ends_slice,
            start_indices=start_indices,
            indices=indices,
            booleans=l_booleans,
        )
    return _generate_indices_equi_join_tupled(
        indices=indices,
        left_index=left_index,
        starts=starts_slice,
        ends=ends_slice,
        start_indices=start_indices,
        right_index=right_index,
        booleans=l_booleans,
    )


@njit(parallel=True, cache=True)
def _numba_equi_join_range_join_non_monotonic(
    left_index: np.ndarray,
    right_index: np.ndarray,
    slice_starts: np.ndarray,
    slice_ends: np.ndarray,
    ge_arr1: np.ndarray,
    ge_arr2: np.ndarray,
    ge_strict: bool,
    row_count: bool,
    tupled: tuple,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Get indices for an equi join
    and a range join
    """
    length = left_index.size
    booleans = np.ones(length, dtype=np.bool_)
    booleans, slice_ends, counts = _ge_or_gt_equi_join(
        source_array=ge_arr1,
        search_array=ge_arr2,
        slice_starts=slice_starts,
        slice_ends=slice_ends,
        strict=ge_strict,
        booleans=booleans,
    )
    if counts == length:
        return None, None

    _indices = np.cumsum(slice_ends - slice_starts)
    start_indices = np.empty(left_index.size, dtype=np.intp)
    start_indices[0] = 0
    start_indices[1:] = _indices[:-1]
    _indices = _indices[-1]
    indices = np.ones(_indices, dtype=np.bool_)
    starts_slice = slice_starts[:]
    ends_slice = slice_ends[:]
    l_booleans = booleans[:]

    l_booleans, indices = _eq_join_tuple_compare(
        booleans=l_booleans,
        indices=indices,
        left_index=left_index,
        start_indices=start_indices,
        starts=starts_slice,
        ends=ends_slice,
        tupled=tupled,
    )

    if not np.any(l_booleans):
        return None, None

    if row_count:
        return _get_row_count_tupled_equi_join(
            left_index=left_index,
            starts=starts_slice,
            ends=ends_slice,
            start_indices=start_indices,
            indices=indices,
            booleans=l_booleans,
        )
    return _generate_indices_equi_join_tupled(
        indices=indices,
        left_index=left_index,
        starts=starts_slice,
        ends=ends_slice,
        start_indices=start_indices,
        right_index=right_index,
        booleans=l_booleans,
    )


@njit(nogil=True)
def _not_montonic_multi_unroll(
    left_indices: np.ndarray,
    right_indices: np.ndarray,
    indices: np.ndarray,
    l_booleans: np.ndarray,
    left_index: np.ndarray,
    counts: np.ndarray,
    tupled: tuple,
):
    """
    Get matching locations for non equi join
    """

    start_indices = np.empty(left_index.size, dtype=np.intp)
    start_indices[0] = 0
    start_indices[1:] = counts.cumsum()[:-1]
    for _tuple in literal_unroll(tupled):
        left_arr = _tuple[0]
        right_arr = _tuple[1]
        op = _tuple[2]
        for n in range(left_index.size):
            _n = np.uintp(n)
            if (not l_booleans[_n]) | (not counts[_n]):
                l_booleans[_n] = False
                continue
            counter = 0
            nn = left_indices[_n]
            _nn = np.uintp(nn)
            size = counts[_n]
            start_index = start_indices[_n]
            left_val = left_arr[_nn]
            for ind in range(start_index, start_index + size):
                _ind = np.uintp(ind)
                nnn = right_indices[_ind]
                _nnn = np.uintp(nnn)
                right_val = right_arr[_nnn]
                boolean = _compare(left_val, right_val, op)
                boolean = indices[_ind] & boolean
                indices[_ind] = boolean
                counter += np.intp(boolean)
            boolean_ = counter > 0
            l_booleans[_n] &= boolean_
    return l_booleans, start_indices, indices


@njit(cache=True, parallel=False, nogil=True)
def _numba_non_equi_join_not_monotonic_keep_all(
    tupled: tuple,
    left_index: np.ndarray,
    right_index: np.ndarray,
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    load_factor: int,
    starts: np.ndarray,
    row_count: bool,
) -> tuple:
    """
    Get indices if there are more than two join conditions
    """
    left_indices, right_indices, counts = (
        _numba_non_equi_join_not_monotonic_keep_all_indices(
            left_regions=left_regions,
            right_regions=right_regions,
            maxxes=maxxes,
            lengths=lengths,
            sorted_array=sorted_array,
            positions_array=positions_array,
            starts=starts,
            load_factor=load_factor,
        )
    )
    if left_indices is None:
        return None, None
    indices = np.ones(right_indices.size, dtype=np.bool_)
    l_booleans = np.ones(left_index.size, dtype=np.bool_)
    l_booleans, start_indices, indices = _not_montonic_multi_unroll(
        left_indices=left_indices,
        right_indices=right_indices,
        left_index=left_index,
        counts=counts,
        tupled=tupled,
        indices=indices,
        l_booleans=l_booleans,
    )
    if not np.any(l_booleans):
        return None, None
    if row_count:
        total = l_booleans.sum()
        r_counts = np.zeros(total, dtype=np.intp)
        left_indexes = np.empty(total, dtype=np.intp)
        indexer = 0
        for n in range(l_booleans.size):
            _n = np.uintp(n)
            if not l_booleans[_n]:
                continue
            size = counts[_n]
            start_index = start_indices[_n]
            counter = 0
            for ind in range(start_index, start_index + size):
                _ind = np.uintp(ind)
                if not indices[_ind]:
                    continue
                counter += 1
            nn = left_indices[_n]
            _nn = np.uintp(nn)
            left_indexes[indexer] = left_index[_nn]
            r_counts[indexer] = counter
            indexer += 1
        return left_indexes, r_counts
    total = indices.sum()
    left_indexes = np.empty(total, dtype=np.intp)
    right_indexes = np.empty(total, dtype=np.intp)
    indexer = 0
    counter = 0
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        nn = left_indices[_n]
        _nn = np.uintp(nn)
        size = counts[_n]
        start_index = start_indices[_n]
        left_val = left_index[_nn]
        for ind in range(start_index, start_index + size):
            _ind = np.uintp(ind)
            boolean = indices[_ind]
            if not boolean:
                continue
            nnn = right_indices[_ind]
            _nnn = np.uintp(nnn)
            right_val = right_index[_nnn]
            _indexer = np.uintp(indexer)
            left_indexes[_indexer] = left_val
            right_indexes[_indexer] = right_val
            indexer += 1
            if indexer == total:
                counter = 1
                break
        if counter == 1:
            break
    return left_indexes, right_indexes


@njit(cache=True, parallel=False, nogil=True)
def _numba_non_equi_join_not_monotonic_keep_first(
    tupled: bool,
    left_index: np.ndarray,
    right_index: np.ndarray,
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    load_factor: int,
    starts: np.ndarray,
) -> tuple:
    """
    Get indices if there are more than two join conditions
    """
    left_indices, right_indices, counts = (
        _numba_non_equi_join_not_monotonic_keep_all_indices(
            left_regions=left_regions,
            right_regions=right_regions,
            maxxes=maxxes,
            lengths=lengths,
            sorted_array=sorted_array,
            positions_array=positions_array,
            starts=starts,
            load_factor=load_factor,
        )
    )
    if left_indices is None:
        return None, None
    indices = np.ones(right_indices.size, dtype=np.bool_)
    l_booleans = np.ones(left_index.size, dtype=np.bool_)
    l_booleans, start_indices, indices = _not_montonic_multi_unroll(
        left_indices=left_indices,
        right_indices=right_indices,
        left_index=left_index,
        counts=counts,
        tupled=tupled,
        indices=indices,
        l_booleans=l_booleans,
    )
    total = l_booleans.sum()
    if not total:
        return None, None
    left_indexes = np.empty(total, dtype=np.intp)
    right_indexes = np.empty(total, dtype=np.intp)
    indexer = 0
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        nn = left_indices[_n]
        _nn = np.uintp(nn)
        size = counts[_n]
        start_index = start_indices[_n]
        left_val = left_index[_nn]
        base = -1
        for ind in range(start_index, start_index + size):
            _ind = np.uintp(ind)
            boolean = indices[_ind]
            if not boolean:
                continue
            nnn = right_indices[_ind]
            _nnn = np.uintp(nnn)
            right_val = right_index[_nnn]
            if (base == -1) | (right_val < base):
                base = right_val
        _indexer = np.uintp(indexer)
        left_indexes[_indexer] = left_val
        right_indexes[_indexer] = base
        indexer += 1
    return left_indexes, right_indexes


@njit(cache=True, parallel=False, nogil=True)
def _numba_non_equi_join_not_monotonic_keep_last(
    tupled: tuple,
    left_index: np.ndarray,
    right_index: np.ndarray,
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    load_factor: int,
    starts: np.ndarray,
) -> tuple:
    """
    Get indices if there are more than two join conditions
    """
    left_indices, right_indices, counts = (
        _numba_non_equi_join_not_monotonic_keep_all_indices(
            left_regions=left_regions,
            right_regions=right_regions,
            maxxes=maxxes,
            lengths=lengths,
            sorted_array=sorted_array,
            positions_array=positions_array,
            starts=starts,
            load_factor=load_factor,
        )
    )
    if left_indices is None:
        return None, None
    indices = np.ones(right_indices.size, dtype=np.bool_)
    l_booleans = np.ones(left_index.size, dtype=np.bool_)
    l_booleans, start_indices, indices = _not_montonic_multi_unroll(
        left_indices=left_indices,
        right_indices=right_indices,
        left_index=left_index,
        counts=counts,
        tupled=tupled,
        indices=indices,
        l_booleans=l_booleans,
    )

    total = l_booleans.sum()
    if not total:
        return None, None
    left_indexes = np.empty(total, dtype=np.intp)
    right_indexes = np.empty(total, dtype=np.intp)
    indexer = 0
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        nn = left_indices[_n]
        _nn = np.uintp(nn)
        size = counts[_n]
        start_index = start_indices[_n]
        left_val = left_index[_nn]
        base = np.inf
        for ind in range(start_index, start_index + size):
            _ind = np.uintp(ind)
            boolean = indices[_ind]
            if not boolean:
                continue
            nnn = right_indices[_ind]
            _nnn = np.uintp(nnn)
            right_val = right_index[_nnn]
            if (base == np.inf) | (right_val > base):
                base = right_val
        _indexer = np.uintp(indexer)
        left_indexes[_indexer] = left_val
        right_indexes[_indexer] = base
        indexer += 1
    return left_indexes, right_indexes


@njit(inline="always")
def compare_values(left_val, right_val, op):
    if op == 0:
        return left_val > right_val
    if op == 1:
        return left_val >= right_val
    if op == 2:
        return left_val < right_val
    if op == 3:
        return left_val <= right_val
    if op == 4:
        return left_val != right_val
    return left_val == right_val


def _compare(x, y, op):
    if (
        (is_numeric_dtype(x) and is_numeric_dtype(y))
        or (is_datetime64_dtype(x) and is_datetime64_dtype(y))
        or (is_timedelta64_dtype(x) and is_timedelta64_dtype(y))
    ):
        return compare_values(x, y, op)


accepted_types = (
    types.NPDatetime,
    types.Integer,
    types.Float,
    types.NPTimedelta,
)


@overload(_compare)
def _numba_compare(x, y, op):
    if (
        isinstance(x, accepted_types)
        and isinstance(y, accepted_types)
        and isinstance(op, types.Integer)
    ):

        def impl(x, y, op):
            return compare_values(x, y, op)

        return impl
    else:
        raise TypeError("Unsupported Type")


@njit(cache=True, parallel=True)
def _range_join_sorted_dual_keep_all(
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    start_indices: np.ndarray,
    left_indices: np.ndarray,
    right_indices: np.ndarray,
):
    """
    Get indices for a dual non equi join
    """
    for ind in prange(left_index.size):
        _ind = np.uintp(ind)
        start = starts[_ind]
        end = ends[_ind]
        indexer = start_indices[_ind]
        lindex = left_index[_ind]
        for num in range(start, end):
            _num = np.uintp(num)
            rindex = right_index[_num]
            _indexer = np.uintp(indexer)
            left_indices[_indexer] = lindex
            right_indices[_indexer] = rindex
            indexer += 1
    return left_indices, right_indices


@njit(cache=True, parallel=True)
def _numba_range_join_sorted_keep_first_or_last_dual(
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    left_indices: np.ndarray,
    right_indices: np.ndarray,
    position: bool,
):
    """
    Get indices for a non equi join
    """
    for ind in prange(left_index.size):
        _ind = np.uintp(ind)
        start = starts[_ind]
        end = ends[_ind]
        lindex = left_index[_ind]
        base_index = right_index[np.uintp(start)]
        for num in range(start, end):
            _num = np.uintp(num)
            rindex = right_index[_num]
            keep_first = rindex < base_index
            keep_first = position & keep_first
            keep_last = rindex > base_index
            keep_last = (not position) & keep_last
            boolean = keep_first | keep_last
            if boolean:
                base_index = rindex
        left_indices[_ind] = lindex
        right_indices[_ind] = base_index
    return left_indices, right_indices


@njit(nogil=True)
def _monotonic_increasing_unroll(
    left_index: np.ndarray,
    right_index: np.ndarray,
    tupled: tuple,
    starts: np.ndarray,
    start_indices: np.ndarray,
    indices: np.ndarray,
):
    """
    Compute matching locations for multiple conditions
    """
    l_booleans = np.ones(left_index.size, dtype=np.bool_)
    end = right_index.size
    for _tuple in literal_unroll(tupled):
        left_arr = _tuple[0]
        right_arr = _tuple[1]
        op = _tuple[2]
        for n in range(left_index.size):
            _n = np.uintp(n)
            if not l_booleans[_n]:
                continue
            start = starts[_n]
            pos = 0
            counter = 0
            left_val = left_arr[_n]
            ind = start_indices[_n]
            for nn in range(start, end):
                _nn = np.uintp(nn)
                right_val = right_arr[_nn]
                boolean = _compare(left_val, right_val, op)
                _ind = np.uintp(ind + pos)
                # pos should always increment
                # no matter what happens
                # with the conditionals below
                pos += 1
                boolean = indices[_ind] & boolean
                indices[_ind] = boolean
                counter += np.intp(boolean)
            boolean_ = counter > 0
            l_booleans[_n] &= boolean_
    return l_booleans, indices


@njit(cache=True, parallel=False, nogil=True)
def _numba_non_equi_join_monotonic_increasing_keep_all(
    tupled: tuple,
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    indices: np.ndarray,
    start_indices: np.ndarray,
    row_count: bool,
) -> tuple:
    """
    Get indices if there are more than two join conditions,
    and a range join, sorted on both right columns, exists.
    """
    l_booleans, indices = _monotonic_increasing_unroll(
        left_index=left_index,
        right_index=right_index,
        tupled=tupled,
        starts=starts,
        start_indices=start_indices,
        indices=indices,
    )
    if not np.any(l_booleans):
        return None, None
    if row_count:
        counts = np.zeros(left_index.size, dtype=np.intp)
        end = right_index.size
        for n in range(left_index.size):
            _n = np.uintp(n)
            if not l_booleans[_n]:
                continue
            start = starts[_n]
            pos = 0
            counter = 0
            ind = start_indices[_n]
            l_index = left_index[_n]
            for nn in range(start, end):
                _ind = np.uintp(ind + pos)
                # pos should always increment
                # no matter what happens
                # with the condition below
                pos += 1
                if not indices[_ind]:
                    continue
                counter += 1
            counts[_n] = counter
        return left_index, counts
    total = indices.sum()
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    indexer = 0
    end = right_index.size
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        start = starts[_n]
        pos = 0
        counter = 0
        ind = start_indices[_n]
        l_index = left_index[_n]
        for nn in range(start, end):
            _ind = np.uintp(ind + pos)
            # pos should always increment
            # no matter what happens
            # with the condition below
            pos += 1
            if not indices[_ind]:
                continue
            _nn = np.uintp(nn)
            _indexer = np.uintp(indexer)
            left_indices[_indexer] = l_index
            right_indices[_indexer] = right_index[_nn]
            indexer += 1
            if indexer == total:
                counter = 1
                break
        if counter == 1:
            break
    return left_indices, right_indices


@njit(cache=True, parallel=False, nogil=True)
def _numba_non_equi_join_monotonic_increasing_keep_first_or_last(
    tupled: tuple,
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    indices: np.ndarray,
    start_indices: np.ndarray,
    position: bool,
) -> tuple:
    """
    Get earliest/latest indices if there are more than two join conditions,
    and a range join, sorted on both right columns, exists.
    """
    l_booleans, indices = _monotonic_increasing_unroll(
        left_index=left_index,
        right_index=right_index,
        tupled=tupled,
        starts=starts,
        start_indices=start_indices,
        indices=indices,
    )

    total = l_booleans.sum()
    if not total:
        return None, None
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    indexer = 0
    end = right_index.size
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        start = starts[_n]
        pos = 0
        ind = start_indices[_n]
        l_index = left_index[_n]
        base = -1
        for nn in range(start, end):
            _ind = np.uintp(ind + pos)
            # pos should always increment
            # no matter what happens
            # with the condition below
            pos += 1
            if not indices[_ind]:
                continue
            _nn = np.uintp(nn)
            value = right_index[_nn]
            keep_first = value < base
            keep_first = position & keep_first
            keep_last = value > base
            keep_last = (not position) & keep_last
            boolean = keep_first | keep_last
            boolean = (base == -1) | boolean
            if boolean:
                base = value
        _indexer = np.uintp(indexer)
        left_indices[_indexer] = l_index
        right_indices[_indexer] = base
        indexer += 1
    return left_indices, right_indices


@njit(nogil=True)
def _range_join_sorted_unroll(
    tupled: tuple,
    left_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    start_indices: np.ndarray,
    indices: np.ndarray,
):
    """
    Get matching locations for multiple conditions for a sorted range join
    """
    l_booleans = np.ones(left_index.size, dtype=np.bool_)
    for _tuple in literal_unroll(tupled):
        left_arr = _tuple[0]
        right_arr = _tuple[1]
        op = _tuple[2]
        for n in range(left_index.size):
            _n = np.uintp(n)
            if not l_booleans[_n]:
                continue
            start = starts[_n]
            end = ends[_n]
            pos = 0
            counter = 0
            ind = start_indices[_n]
            for nn in range(start, end):
                _nn = np.uintp(nn)
                left_val = left_arr[_n]
                right_val = right_arr[_nn]
                boolean = _compare(left_val, right_val, op)
                _ind = np.uintp(ind + pos)
                # pos should always increment
                # no matter what happens
                # with the conditionals below
                pos += 1
                boolean = indices[_ind] & boolean
                indices[_ind] = boolean
                counter += np.intp(boolean)
            boolean_ = counter > 0
            l_booleans[_n] &= boolean_
    return l_booleans, indices


@njit(cache=True, parallel=False, nogil=True)
def _range_join_sorted_multiple_keep_all(
    tupled: tuple,
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    indices: np.ndarray,
    start_indices: np.ndarray,
    row_count: bool,
) -> tuple:
    """
    Get indices if there are more than two join conditions,
    and a range join, sorted on both right columns, exists.
    """
    l_booleans, indices = _range_join_sorted_unroll(
        tupled=tupled,
        left_index=left_index,
        starts=starts,
        ends=ends,
        start_indices=start_indices,
        indices=indices,
    )
    if not np.any(l_booleans):
        return None, None
    if row_count:
        counts = np.zeros(left_index.size, dtype=np.intp)
        for n in range(left_index.size):
            _n = np.uintp(n)
            if not l_booleans[_n]:
                continue
            start = starts[_n]
            end = ends[_n]
            pos = 0
            counter = 0
            ind = start_indices[_n]
            for _ in range(start, end):
                _ind = np.uintp(ind + pos)
                # pos should always increment
                # no matter what happens
                # with the condition below
                pos += 1
                if not indices[_ind]:
                    continue
                counter += 1
            counts[_n] = counter
        return left_index, counts
    total = indices.sum()
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    indexer = 0
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        start = starts[_n]
        end = ends[_n]
        pos = 0
        counter = 0
        ind = start_indices[_n]
        l_index = left_index[_n]
        for nn in range(start, end):
            _ind = np.uintp(ind + pos)
            # pos should always increment
            # no matter what happens
            # with the condition below
            pos += 1
            if not indices[_ind]:
                continue
            _nn = np.uintp(nn)
            _indexer = np.uintp(indexer)
            left_indices[_indexer] = l_index
            right_indices[_indexer] = right_index[_nn]
            indexer += 1
            if indexer == total:
                counter = 1
                break
        if counter == 1:
            break
    return left_indices, right_indices


@njit(cache=True, parallel=False, nogil=True)
def _range_join_sorted_multiple_keep_first_or_last(
    tupled: tuple,
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
    indices: np.ndarray,
    start_indices: np.ndarray,
    position: bool,
) -> tuple:
    """
    Get earliest/latest indices if there are more than two join conditions,
    and a range join, sorted on both right columns, exists.
    """
    l_booleans, indices = _range_join_sorted_unroll(
        tupled=tupled,
        left_index=left_index,
        starts=starts,
        ends=ends,
        start_indices=start_indices,
        indices=indices,
    )
    total = l_booleans.sum()
    if not total:
        return None, None
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    indexer = 0
    for n in range(left_index.size):
        _n = np.uintp(n)
        if not l_booleans[_n]:
            continue
        start = starts[_n]
        end = ends[_n]
        pos = 0
        ind = start_indices[_n]
        l_index = left_index[_n]
        base = -1
        for nn in range(start, end):
            _ind = np.uintp(ind + pos)
            # pos should always increment
            # no matter what happens
            # with the condition below
            pos += 1
            if not indices[_ind]:
                continue
            _nn = np.uintp(nn)
            value = right_index[_nn]
            keep_first = value < base
            keep_first = position & keep_first
            keep_last = value > base
            keep_last = (not position) & keep_last
            boolean = keep_first | keep_last
            boolean = (base == -1) | boolean
            if boolean:
                base = value
        _indexer = np.uintp(indexer)
        left_indices[_indexer] = l_index
        right_indices[_indexer] = base
        indexer += 1
    return left_indices, right_indices


@njit(cache=True, parallel=True)
def _numba_non_equi_join_monotonic_increasing_keep_first_or_last_dual(
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    left_indices: np.ndarray,
    right_indices: np.ndarray,
    position: bool,
):
    """
    Get indices for a non equi join
    """
    end = right_index.size
    for ind in prange(left_index.size):
        _ind = np.uintp(ind)
        start = starts[_ind]
        lindex = left_index[_ind]
        base_index = right_index[np.uintp(start)]
        for num in range(start, end):
            _num = np.uintp(num)
            rindex = right_index[_num]
            keep_first = rindex < base_index
            keep_first = position & keep_first
            keep_last = rindex > base_index
            keep_last = (not position) & keep_last
            boolean = keep_first | keep_last
            if boolean:
                base_index = rindex
        left_indices[_ind] = lindex
        right_indices[_ind] = base_index
    return left_indices, right_indices


@njit(cache=True, parallel=True)
def _numba_non_equi_join_monotonic_increasing_keep_all_dual(
    left_index: np.ndarray,
    right_index: np.ndarray,
    starts: np.ndarray,
    start_indices: np.ndarray,
    left_indices: np.ndarray,
    right_indices: np.ndarray,
):
    """
    Get indices for a non equi join
    """
    end = right_index.size
    for ind in prange(left_index.size):
        _ind = np.uintp(ind)
        start = starts[_ind]
        indexer = start_indices[_ind]
        lindex = left_index[_ind]
        for num in range(start, end):
            _num = np.uintp(num)
            rindex = right_index[_num]
            _indexer = np.uintp(indexer)
            left_indices[_indexer] = lindex
            right_indices[_indexer] = rindex
            indexer += 1
    return left_indices, right_indices


@njit(nogil=True)
def _numba_non_equi_join_not_monotonic_keep_all_indices(
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    starts: np.ndarray,
    load_factor: int,
):
    """
    Get indices for non-equi join,
    where the right regions are not monotonic
    """
    # first pass - get actual length
    length = left_regions.size
    end = right_regions.size
    end -= 1
    # add the last region
    # no need to have this checked within an if-else statement
    # in the for loop below
    region = right_regions[np.uintp(end)]
    sorted_array[0, 0] = region
    positions_array[0, 0] = end
    # keep track of the maxxes array
    # how many cells have actual values?
    maxxes_counter = 1
    maxxes[0] = region
    lengths[0] = 1
    r_count = 0
    total = 0
    l_booleans = np.zeros(length, dtype=np.bool_)
    for indexer in range(length - 1, -1, -1):
        _indexer = np.uintp(indexer)
        start = starts[_indexer]
        for num in range(start, end):
            _num = np.uintp(num)
            region = right_regions[_num]
            arr = maxxes[:maxxes_counter]
            if region > arr[-1]:
                # it is larger than the max in the maxxes array
                # shove it into the last column
                posn = maxxes_counter - 1
                posn_ = np.uintp(posn)
                len_arr = lengths[posn_]
                len_arr_ = np.uintp(len_arr)
                sorted_array[len_arr_, posn_] = region
                # we dont need to compute positions in the first run?
                positions_array[len_arr_, posn_] = num
                maxxes[posn_] = region
                lengths[posn_] += 1
            else:
                posn = _numba_less_than_base(arr=arr, value=region)
                sorted_array, positions_array, lengths, maxxes = _numba_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    maxxes=maxxes,
                    lengths=lengths,
                    region=region,
                    posn=posn,
                    num=num,
                )
            r_count += 1
            posn_ = np.uintp(posn)
            # have we exceeded the size of this column?
            # do we need to trim and move data to other columns?
            check = (lengths[posn_] == (load_factor * 2)) & (
                r_count < right_regions.size
            )
            if check:
                (
                    sorted_array,
                    positions_array,
                    lengths,
                    maxxes,
                    maxxes_counter,
                ) = _expand_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    lengths=lengths,
                    maxxes=maxxes,
                    posn=posn,
                    maxxes_counter=maxxes_counter,
                    load_factor=load_factor,
                )
        # now we do a binary search
        # for left region in right region
        # 1. find the position in maxxes
        # - this indicates which column in sorted_arrays contains our region
        # 2. search in the specific region for the positions
        # where left_region <= right_region
        l_region = left_regions[_indexer]
        arr = maxxes[:maxxes_counter]
        if l_region > arr[-1]:
            end = start
            continue
        posn = _numba_less_than_base(arr=arr, value=l_region)
        posn_ = np.uintp(posn)
        len_arr = lengths[posn_]
        arr = sorted_array[:len_arr, posn_]
        _posn = _numba_less_than_base(arr=arr, value=l_region)
        difference = len_arr - _posn
        total += difference
        # step into the remaining columns
        for ind in range(posn + 1, maxxes_counter):
            ind_ = np.uintp(ind)
            len_arr = lengths[ind_]
            difference += len_arr
            total += len_arr
        l_booleans[_indexer] = True
        end = start
    if total == 0:
        return None, None, None
    # second pass - fill arrays with indices
    length = left_regions.size
    end = right_regions.size
    end -= 1
    region = right_regions[np.uintp(end)]
    sorted_array[0, 0] = region
    positions_array[0, 0] = end
    maxxes_counter = 1
    maxxes[0] = region
    lengths[0] = 1
    r_count = 0
    left_counts = np.zeros(length, dtype=np.intp)
    left_indices = np.empty(length, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    begin = 0
    l_indexer = 0
    for indexer in range(length - 1, -1, -1):
        _indexer = np.uintp(indexer)
        if not l_booleans[_indexer]:
            l_indexer += 1
            continue
        start = starts[_indexer]
        for num in range(start, end):
            _num = np.uintp(num)
            region = right_regions[_num]
            arr = maxxes[:maxxes_counter]
            if region > arr[-1]:
                posn = maxxes_counter - 1
                posn_ = np.uintp(posn)
                len_arr = lengths[posn_]
                len_arr_ = np.uintp(len_arr)
                sorted_array[len_arr_, posn_] = region
                positions_array[len_arr_, posn_] = num
                maxxes[posn_] = region
                lengths[posn_] += 1
            else:
                posn = _numba_less_than_base(arr=arr, value=region)
                sorted_array, positions_array, lengths, maxxes = _numba_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    maxxes=maxxes,
                    lengths=lengths,
                    region=region,
                    posn=posn,
                    num=num,
                )
            r_count += 1
            posn_ = np.uintp(posn)
            # have we reached the max size of this column?
            # do we need to trim and move data to other columns?
            check = (lengths[posn_] == (load_factor * 2)) & (
                r_count < right_regions.size
            )
            if check:
                (
                    sorted_array,
                    positions_array,
                    lengths,
                    maxxes,
                    maxxes_counter,
                ) = _expand_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    lengths=lengths,
                    maxxes=maxxes,
                    posn=posn,
                    maxxes_counter=maxxes_counter,
                    load_factor=load_factor,
                )
        # now we do a binary search
        # for left region in right region
        l_region = left_regions[_indexer]
        arr = maxxes[:maxxes_counter]
        left_indices[np.uintp(l_indexer)] = indexer
        if l_region > arr[-1]:
            end = start
            l_indexer += 1
            continue
        counter = 0
        posn = _numba_less_than_base(arr=arr, value=l_region)
        posn_ = np.uintp(posn)
        len_arr = lengths[posn_]
        arr = sorted_array[:len_arr, posn_]
        _posn = _numba_less_than_base(arr=arr, value=l_region)
        for ind in range(_posn, len_arr):
            ind_ = np.uintp(ind)
            begin_ = np.uintp(begin)
            r_pos = positions_array[ind_, posn_]
            right_indices[begin_] = r_pos
            begin += 1
            counter += 1
        for ind in range(posn + 1, maxxes_counter):
            ind_ = np.uintp(ind)
            len_arr = lengths[ind_]
            for num in range(len_arr):
                _num = np.uintp(num)
                begin_ = np.uintp(begin)
                r_pos = positions_array[_num, ind_]
                right_indices[begin_] = r_pos
                begin += 1
                counter += 1
        left_counts[l_indexer] = counter
        left_indices[l_indexer] = indexer
        l_indexer += 1
        end = start
    return left_indices, right_indices, left_counts


@njit(cache=True)
def _numba_non_equi_join_not_monotonic_dual_keep_all(
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    left_index: np.ndarray,
    right_index: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    starts: np.ndarray,
    load_factor: int,
    row_count: bool,
):
    """
    Get indices for non-equi join,
    where the right regions are not monotonic
    """
    # first pass - get actual length
    length = left_index.size
    end = right_index.size
    end -= 1
    # add the last region
    # no need to have this checked within an if-else statement
    # in the for loop below
    region = right_regions[np.uintp(end)]
    sorted_array[0, 0] = region
    positions_array[0, 0] = end
    # keep track of the maxxes array
    # how many cells have actual values?
    maxxes_counter = 1
    maxxes[0] = region
    lengths[0] = 1
    r_count = 0
    total = 0
    l_booleans = np.zeros(length, dtype=np.bool_)
    if row_count:
        counts = np.zeros(length, dtype=np.intp)
    for indexer in range(length - 1, -1, -1):
        _indexer = np.uintp(indexer)
        start = starts[_indexer]
        for num in range(start, end):
            _num = np.uintp(num)
            region = right_regions[_num]
            arr = maxxes[:maxxes_counter]
            if region > arr[-1]:
                # it is larger than the max in the maxxes array
                # shove it into the last column
                posn = maxxes_counter - 1
                posn_ = np.uintp(posn)
                len_arr = lengths[posn_]
                len_arr_ = np.uintp(len_arr)
                sorted_array[len_arr_, posn_] = region
                # we dont need to compute positions in the first run?
                positions_array[len_arr_, posn_] = num
                maxxes[posn_] = region
                lengths[posn_] += 1
            else:
                posn = _numba_less_than_base(arr=arr, value=region)
                sorted_array, positions_array, lengths, maxxes = _numba_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    maxxes=maxxes,
                    lengths=lengths,
                    region=region,
                    posn=posn,
                    num=num,
                )
            r_count += 1
            posn_ = np.uintp(posn)
            # have we exceeded the size of this column?
            # do we need to trim and move data to other columns?
            check = (lengths[posn_] == (load_factor * 2)) & (r_count < right_index.size)
            if check:
                (
                    sorted_array,
                    positions_array,
                    lengths,
                    maxxes,
                    maxxes_counter,
                ) = _expand_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    lengths=lengths,
                    maxxes=maxxes,
                    posn=posn,
                    maxxes_counter=maxxes_counter,
                    load_factor=load_factor,
                )
        # now we do a binary search
        # for left region in right region
        # 1. find the position in maxxes
        # - this indicates which column in sorted_arrays contains our region
        # 2. search in the specific region for the positions
        # where left_region <= right_region
        l_region = left_regions[_indexer]
        arr = maxxes[:maxxes_counter]
        if l_region > arr[-1]:
            end = start
            continue
        posn = _numba_less_than_base(arr=arr, value=l_region)
        posn_ = np.uintp(posn)
        len_arr = lengths[posn_]
        arr = sorted_array[:len_arr, posn_]
        _posn = _numba_less_than_base(arr=arr, value=l_region)
        l_count = len_arr - _posn
        total += len_arr - _posn
        # step into the remaining columns
        for ind in range(posn + 1, maxxes_counter):
            ind_ = np.uintp(ind)
            len_arr = lengths[ind_]
            l_count += len_arr
            total += len_arr
        if row_count:
            counts[_indexer] = l_count
        l_booleans[_indexer] = True
        end = start
    if total == 0:
        return None, None
    if row_count:
        return left_index, counts
    # second pass - fill arrays with indices
    length = left_index.size
    end = right_index.size
    end -= 1
    region = right_regions[np.uintp(end)]
    sorted_array[0, 0] = region
    positions_array[0, 0] = end
    maxxes_counter = 1
    maxxes[0] = region
    lengths[0] = 1
    r_count = 0
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    begin = 0
    for indexer in range(length - 1, -1, -1):
        _indexer = np.uintp(indexer)
        if not l_booleans[_indexer]:
            continue
        start = starts[_indexer]
        for num in range(start, end):
            _num = np.uintp(num)
            region = right_regions[_num]
            arr = maxxes[:maxxes_counter]
            if region > arr[-1]:
                posn = maxxes_counter - 1
                posn_ = np.uintp(posn)
                len_arr = lengths[posn_]
                len_arr_ = np.uintp(len_arr)
                sorted_array[len_arr_, posn_] = region
                positions_array[len_arr_, posn_] = num
                maxxes[posn_] = region
                lengths[posn_] += 1
            else:
                posn = _numba_less_than_base(arr=arr, value=region)
                sorted_array, positions_array, lengths, maxxes = _numba_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    maxxes=maxxes,
                    lengths=lengths,
                    region=region,
                    posn=posn,
                    num=num,
                )
            r_count += 1
            posn_ = np.uintp(posn)
            # have we reached the max size of this column?
            # do we need to trim and move data to other columns?
            check = (lengths[posn_] == (load_factor * 2)) & (r_count < right_index.size)
            if check:
                (
                    sorted_array,
                    positions_array,
                    lengths,
                    maxxes,
                    maxxes_counter,
                ) = _expand_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    lengths=lengths,
                    maxxes=maxxes,
                    posn=posn,
                    maxxes_counter=maxxes_counter,
                    load_factor=load_factor,
                )
        # now we do a binary search
        # for left region in right region
        l_region = left_regions[_indexer]
        arr = maxxes[:maxxes_counter]
        if l_region > arr[-1]:
            end = start
            continue
        posn = _numba_less_than_base(arr=arr, value=l_region)
        posn_ = np.uintp(posn)
        len_arr = lengths[posn_]
        arr = sorted_array[:len_arr, posn_]
        _posn = _numba_less_than_base(arr=arr, value=l_region)
        l_index = left_index[_indexer]
        for ind in range(_posn, len_arr):
            ind_ = np.uintp(ind)
            begin_ = np.uintp(begin)
            r_pos = positions_array[ind_, posn_]
            r_pos = np.uintp(r_pos)
            r_index = right_index[r_pos]
            left_indices[begin_] = l_index
            right_indices[begin_] = r_index
            begin += 1
        for ind in range(posn + 1, maxxes_counter):
            ind_ = np.uintp(ind)
            len_arr = lengths[ind_]
            for num in range(len_arr):
                _num = np.uintp(num)
                begin_ = np.uintp(begin)
                left_indices[begin_] = l_index
                r_pos = positions_array[_num, ind_]
                r_pos = np.uintp(r_pos)
                r_index = right_index[r_pos]
                right_indices[begin_] = r_index
                begin += 1
        end = start
    return left_indices, right_indices


@njit(cache=True)
def _numba_non_equi_join_not_monotonic_dual_keep_first(
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    left_index: np.ndarray,
    right_index: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    starts: np.ndarray,
    load_factor: int,
):
    """
    Get indices for non-equi join - first match
    """
    # first pass - get the actual length
    length = left_index.size
    end = right_index.size
    end -= 1
    region = right_regions[np.uintp(end)]
    sorted_array[0, 0] = region
    positions_array[0, 0] = end
    maxxes_counter = 1
    maxxes[0] = region
    lengths[0] = 1
    r_count = 0
    total = 0
    l_booleans = np.zeros(length, dtype=np.bool_)
    r_indices = np.empty(length, dtype=np.intp)
    for indexer in range(length - 1, -1, -1):
        _indexer = np.uintp(indexer)
        start = starts[_indexer]
        for num in range(start, end):
            _num = np.uintp(num)
            region = right_regions[_num]
            arr = maxxes[:maxxes_counter]
            if region > arr[-1]:
                posn = maxxes_counter - 1
                posn_ = np.uintp(posn)
                len_arr = lengths[posn_]
                len_arr_ = np.uintp(len_arr)
                sorted_array[len_arr_, posn_] = region
                positions_array[len_arr_, posn_] = num
                maxxes[posn_] = region
                lengths[posn_] += 1
            else:
                posn = _numba_less_than_base(arr=arr, value=region)
                sorted_array, positions_array, lengths, maxxes = _numba_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    maxxes=maxxes,
                    lengths=lengths,
                    region=region,
                    posn=posn,
                    num=num,
                )
            r_count += 1
            posn_ = np.uintp(posn)
            # have we exceeded the size of this column?
            # do we need to trim and move data to other columns?
            check = (lengths[posn_] == (load_factor * 2)) & (r_count < right_index.size)
            if check:
                (
                    sorted_array,
                    positions_array,
                    lengths,
                    maxxes,
                    maxxes_counter,
                ) = _expand_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    lengths=lengths,
                    maxxes=maxxes,
                    posn=posn,
                    maxxes_counter=maxxes_counter,
                    load_factor=load_factor,
                )
        l_region = left_regions[_indexer]
        arr = maxxes[:maxxes_counter]
        if l_region > arr[-1]:
            end = start
            continue
        posn = _numba_less_than_base(arr=arr, value=l_region)
        posn_ = np.uintp(posn)
        len_arr = lengths[posn_]
        arr = sorted_array[:len_arr, posn_]
        _posn = _numba_less_than_base(arr=arr, value=l_region)
        base_index = -1
        for ind in range(_posn, len_arr):
            ind_ = np.uintp(ind)
            r_pos = positions_array[ind_, posn_]
            r_pos = np.uintp(r_pos)
            r_index = right_index[r_pos]
            if (base_index == -1) | (r_index < base_index):
                base_index = r_index
        # step into the remaining columns
        for ind in range(posn + 1, maxxes_counter):
            ind_ = np.uintp(ind)
            len_arr = lengths[ind_]
            # step into the rows for each column
            for num in range(len_arr):
                _num = np.uintp(num)
                r_pos = positions_array[_num, ind_]
                r_pos = np.uintp(r_pos)
                r_index = right_index[r_pos]
                if (base_index == -1) | (r_index < base_index):
                    base_index = r_index
        total += 1
        l_booleans[_indexer] = True
        r_indices[_indexer] = base_index
        end = start
    if total == 0:
        return None, None
    # second pass - fill arrays with indices
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    n = 0
    for ind in range(length):
        _ind = np.uintp(ind)
        if not l_booleans[_ind]:
            continue
        _n = np.uintp(n)
        left_indices[_n] = left_index[_ind]
        right_indices[_n] = r_indices[_ind]
        n += 1
    return left_indices, right_indices


@njit(cache=True)
def _numba_non_equi_join_not_monotonic_dual_keep_last(
    left_regions: np.ndarray,
    right_regions: np.ndarray,
    left_index: np.ndarray,
    right_index: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    starts: np.ndarray,
    load_factor: int,
):
    """
    Get indices for non-equi join - last match
    """
    # first pass - get the actual length
    length = left_index.size
    end = right_index.size
    end -= 1
    region = right_regions[np.uintp(end)]
    sorted_array[0, 0] = region
    positions_array[0, 0] = end
    maxxes_counter = 1
    maxxes[0] = region
    lengths[0] = 1
    r_count = 0
    total = 0
    l_booleans = np.zeros(length, dtype=np.bool_)
    r_indices = np.empty(length, dtype=np.intp)
    for indexer in range(length - 1, -1, -1):
        _indexer = np.uintp(indexer)
        start = starts[_indexer]
        for num in range(start, end):
            _num = np.uintp(num)
            region = right_regions[_num]
            arr = maxxes[:maxxes_counter]
            if region > arr[-1]:
                posn = maxxes_counter - 1
                posn_ = np.uintp(posn)
                len_arr = lengths[posn_]
                len_arr_ = np.uintp(len_arr)
                sorted_array[len_arr_, posn_] = region
                positions_array[len_arr_, posn_] = num
                maxxes[posn_] = region
                lengths[posn_] += 1
            else:
                posn = _numba_less_than_base(arr=arr, value=region)
                sorted_array, positions_array, lengths, maxxes = _numba_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    maxxes=maxxes,
                    lengths=lengths,
                    region=region,
                    posn=posn,
                    num=num,
                )
            r_count += 1
            posn_ = np.uintp(posn)
            # have we exceeded the size of this column?
            # do we need to trim and move data to other columns?
            check = (lengths[posn_] == (load_factor * 2)) & (r_count < right_index.size)
            if check:
                (
                    sorted_array,
                    positions_array,
                    lengths,
                    maxxes,
                    maxxes_counter,
                ) = _expand_sorted_array(
                    sorted_array=sorted_array,
                    positions_array=positions_array,
                    lengths=lengths,
                    maxxes=maxxes,
                    posn=posn,
                    maxxes_counter=maxxes_counter,
                    load_factor=load_factor,
                )
        l_region = left_regions[_indexer]
        arr = maxxes[:maxxes_counter]
        posn = _numba_less_than_base(arr=arr, value=l_region)
        if l_region > arr[-1]:
            end = start
            continue
        posn_ = np.uintp(posn)
        len_arr = lengths[posn_]
        arr = sorted_array[:len_arr, posn_]
        _posn = _numba_less_than_base(arr=arr, value=l_region)
        base_index = np.inf
        for ind in range(_posn, len_arr):
            ind_ = np.uintp(ind)
            r_pos = positions_array[ind_, posn_]
            r_pos = np.uintp(r_pos)
            r_index = right_index[r_pos]
            if (base_index == np.inf) | (r_index > base_index):
                base_index = r_index
        # step into the remaining columns
        for ind in range(posn + 1, maxxes_counter):
            ind_ = np.uintp(ind)
            len_arr = lengths[ind_]
            # step into the rows for each column
            for num in range(len_arr):
                _num = np.uintp(num)
                r_pos = positions_array[_num, ind_]
                r_pos = np.uintp(r_pos)
                r_index = right_index[r_pos]
                if (base_index == np.inf) | (r_index > base_index):
                    base_index = r_index
        total += 1
        l_booleans[_indexer] = True
        r_indices[_indexer] = base_index
        end = start
    if total == 0:
        return None, None
    # second pass - fill arrays with indices
    left_indices = np.empty(total, dtype=np.intp)
    right_indices = np.empty(total, dtype=np.intp)
    n = 0
    for ind in range(length):
        _ind = np.uintp(ind)
        if not l_booleans[_ind]:
            continue
        _n = np.uintp(n)
        left_indices[_n] = left_index[_ind]
        right_indices[_n] = r_indices[_ind]
        n += 1
    return left_indices, right_indices


@njit
def _numba_sorted_array(
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    maxxes: np.ndarray,
    lengths: np.ndarray,
    region: int,
    posn: int,
    num: int,
) -> tuple:
    """
    Adaptation of grantjenk's sortedcontainers.

    Args:
        sorted_array: array of regions to keep in sorted order.
        positions_array: positions of regions in the sorted_array.
        maxxes: array of max values per column in the sorted_array.
        lengths: array of lengths per column in the sorted_array.
        region: integer to insert into sorted_array.
        posn: binary search position of region in maxxes array.
            Determines which column in the sorted_array
            the region will go to.
        num: position of region in right_regions array.
            Inserted into positions_array to keep
            in sync with the region the sorted_array.
    """
    # the sorted array is an adaptation
    # of grantjenks' sortedcontainers
    posn_ = np.uintp(posn)
    len_arr = lengths[posn_]
    # grab the specific column that the region falls into
    arr = sorted_array[:len_arr, posn_]
    # get the insertion position for the region
    insort_posn = _numba_less_than_base(arr=arr, value=region)
    # make space for the region
    # shift downwards before inserting
    # shift in this order to avoid issues with assignment override
    # which could create wrong values
    for ind in range(len_arr - 1, insort_posn - 1, -1):
        ind_ = np.uintp(ind)
        _ind = np.uintp(ind + 1)
        sorted_array[_ind, posn_] = sorted_array[ind_, posn_]
        positions_array[_ind, posn_] = positions_array[ind_, posn_]
    # now we can safely insert the region
    insort = np.uintp(insort_posn)
    sorted_array[insort, posn_] = region
    positions_array[insort, posn_] = num
    # update the length and the maxxes arrays
    lengths[posn_] += 1
    maxxes[posn_] = sorted_array[np.uintp(len_arr), posn_]
    return sorted_array, positions_array, lengths, maxxes


@njit
def _expand_sorted_array(
    sorted_array: np.ndarray,
    positions_array: np.ndarray,
    lengths: np.ndarray,
    maxxes: np.ndarray,
    posn: int,
    maxxes_counter: int,
    load_factor: int,
):
    """
    Expand sorted_array if it exceeds load_factor * 2
    Adapted from grantjenks' sortedcontainers.

    Args:
        sorted_array: array of regions to keep in sorted order.
        positions_array: positions of regions in the sorted_array.
        maxxes: array of max values per column in the sorted_array.
        lengths: array of lengths per column in the sorted_array.
        region: integer to insert into sorted_array.
        posn: binary search position of region in maxxes array.
            Determines which column in the sorted_array
            the region will go to.
        maxxes_counter: keeps a count of the number
            of entries in the maxxes array that have
            actual values.
        num: position of region in right_regions array.
            Inserted into positions_array to keep
            in sync with the region the sorted_array.
    """
    # shift from left+1 to right
    for pos in range(maxxes_counter - 1, posn, -1):
        forward = np.uintp(pos + 1)
        current = np.uintp(pos)
        sorted_array[:, forward] = sorted_array[:, current]
        positions_array[:, forward] = positions_array[:, current]
        maxxes[forward] = maxxes[current]
        lengths[forward] = lengths[current]
    # share half the load from left to left+1
    forward = np.uintp(posn + 1)
    current = np.uintp(posn)
    maxxes[forward] = sorted_array[-1, current]
    lengths[forward] = load_factor
    sorted_array[:load_factor, forward] = sorted_array[load_factor:, current]
    positions_array[:load_factor, forward] = positions_array[load_factor:, current]
    # update the length and maxxes arrays
    lengths[current] = load_factor
    maxxes[current] = sorted_array[np.uintp(load_factor - 1), current]
    maxxes_counter += 1
    return sorted_array, positions_array, lengths, maxxes, maxxes_counter



# helper functions for >/>=
import janitor_rs
import numpy as np
import pandas as pd

from janitor.functions._conditional_join import _binary_search
from janitor.functions._conditional_join._helpers import (
    _convert_array_to_numpy,
    _null_checks_cond_join,
    _sort_if_not_monotonic,
)


def _greater_than_indices(
    left: pd.Series,
    right: pd.Series,
    strict: bool,
    keep: str,
    return_matching_indices: bool,
) -> dict | None:
    """
    Use binary search to get indices where left
    is greater than or equal to right.

    If strict is True, then only indices
    where `left` is greater than
    (but not equal to) `right` are returned.
    """
    empty_array = np.array([], dtype=np.intp)
    outcome = _null_checks_cond_join(series=left)
    if outcome is None:
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    left, _ = outcome
    outcome = _null_checks_cond_join(series=right)
    if outcome is None:
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    right, any_nulls = outcome
    right, right_is_sorted = _sort_if_not_monotonic(series=right)
    left_array = _convert_array_to_numpy(array=left._values)
    right_array = _convert_array_to_numpy(array=right._values)
    if strict:
        outcome = _binary_search._binary_search_gt_first(
            left=left_array, right=right_array, left_index=left.index._values
        )
    else:
        outcome = _binary_search._binary_search_ge_first(
            left=left_array, right=right_array, left_index=left.index._values
        )
    if outcome is None:
        return {
            "left_index": empty_array,
            "right_index": empty_array,
        }
    left_index, search_indices = outcome
    right_index = right.index._values
    if right_is_sorted & (keep == "first"):
        indexer = np.zeros_like(search_indices)
        return {"left_index": left_index, "right_index": right_index[indexer]}
    if right_is_sorted & (keep == "last") & any_nulls:
        return {
            "left_index": left_index,
            "right_index": right_index[search_indices - 1],
        }
    if right_is_sorted & (keep == "last"):
        return {"left_index": left_index, "right_index": search_indices - 1}
    if keep == "first":
        right = [right_index[:ind] for ind in search_indices]
        right = [arr.min() for arr in right]
        return {"left_index": left_index, "right_index": right}
    if keep == "last":
        right = [right_index[:ind] for ind in search_indices]
        right = [arr.max() for arr in right]
        return {"left_index": left_index, "right_index": right}
    if return_matching_indices:
        return dict(
            left_index=left_index,
            right_index=right_index,
            starts=0,
            ends=search_indices,
        )
    right = [right_index[:ind] for ind in search_indices]
    right = np.concatenate(right)
    left = janitor_rs.repeat_index(
        index=left_index,
        counts=search_indices,
        length=search_indices.sum(),
    )
    return {"left_index": left, "right_index": right}



"""Implementation of the `collapse_levels` function."""

import pandas as pd
import pandas_flavor as pf
from pandas.api.types import is_string_dtype

from janitor.utils import check


@pf.register_dataframe_method
def collapse_levels(
    df: pd.DataFrame,
    sep: str = None,
    glue: str = None,
    axis: str = "columns",
) -> pd.DataFrame:
    """Flatten multi-level index/column dataframe to a single level.

    This method does not mutate the original DataFrame.

    Given a DataFrame containing multi-level index/columns, flatten to single-level
    by string-joining the labels in each level.

    After a `groupby` / `aggregate` operation where `.agg()` is passed a
    list of multiple aggregation functions, a multi-level DataFrame is
    returned with the name of the function applied in the second level.

    It is sometimes convenient for later indexing to flatten out this
    multi-level configuration back into a single level. This function does
    this through a simple string-joining of all the names across different
    levels in a single column.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "class": ["bird", "bird", "bird", "mammal", "mammal"],
        ...         "max_speed": [389, 389, 24, 80, 21],
        ...         "type": ["falcon", "falcon", "parrot", "Lion", "Monkey"],
        ...     }
        ... )
        >>> df
            class  max_speed    type
        0    bird        389  falcon
        1    bird        389  falcon
        2    bird         24  parrot
        3  mammal         80    Lion
        4  mammal         21  Monkey
        >>> grouped_df = df.groupby("class")[["max_speed"]].agg(["mean", "median"])
        >>> grouped_df  # doctest: +NORMALIZE_WHITESPACE
                 max_speed
                      mean median
        class
        bird    267.333333  389.0
        mammal   50.500000   50.5
        >>> grouped_df.collapse_levels(sep="_")  # doctest: +NORMALIZE_WHITESPACE
                max_speed_mean  max_speed_median
        class
        bird        267.333333             389.0
        mammal       50.500000              50.5

        Before applying `.collapse_levels`, the `.agg` operation returns a
        multi-level column DataFrame whose columns are `(level 1, level 2)`:

        ```python
        [("max_speed", "mean"), ("max_speed", "median")]
        ```

        `.collapse_levels` then flattens the column MultiIndex into a single
        level index with names:

        ```python
        ["max_speed_mean", "max_speed_median"]
        ```

        For more control, a `glue` specification can be passed,
        where the names of the levels are used to control the output of the
        flattened index:
        >>> (
        ...     grouped_df.rename_axis(
        ...         columns=["column_name", "agg_name"]
        ...     ).collapse_levels(glue="{agg_name}_{column_name}")
        ... )  # doctest: +NORMALIZE_WHITESPACE
                mean_max_speed  median_max_speed
        class
        bird        267.333333             389.0
        mammal       50.500000              50.5

        Note that for `glue` to work, the keyword arguments
        in the glue specification
        should be the names of the levels in the MultiIndex.

    !!! abstract "Version Changed"

        - 0.27.0
            - Added `glue` and `axis` parameters.

    Args:
        df: A pandas DataFrame.
        sep: String separator used to join the column level names.
        glue: A specification on how the column levels should be combined.
            It allows for a more granular composition,
            and serves as an alternative to `sep`.
        axis: Determines whether to collapse the
            levels on the index or columns.

    Returns:
        A pandas DataFrame with single-level column index.
    """  # noqa: E501
    if (sep is not None) and (glue is not None):
        raise ValueError("Only one of sep or glue should be provided.")
    if sep is not None:
        check("sep", sep, [str])
    if glue is not None:
        check("glue", glue, [str])
    check("axis", axis, [str])
    if axis not in {"index", "columns"}:
        raise ValueError("axis argument should be either 'index' or 'columns'.")

    if not isinstance(getattr(df, axis), pd.MultiIndex):
        return df

    # TODO: Pyarrow offers faster string computations
    # future work should take this into consideration,
    # which would require a different route from python's string.join
    # since work is only on the columns
    # Use copy() to preserve MultiIndex structure (df[:] can lose it)
    # Pandas creates a new Index altogether, so original dataframe is not modified
    df = df.copy()
    new_index = getattr(df, axis)
    # Re-check after copy (structure should be preserved with copy())
    if not isinstance(new_index, pd.MultiIndex):
        return df
    if glue is not None:
        new_index = [dict(zip(new_index.names, entry)) for entry in new_index]
        new_index = [glue.format_map(mapping) for mapping in new_index]
        setattr(df, axis, new_index)
        return df
    sep = "_" if sep is None else sep
    # Access levels safely (works with latest pandas)
    # Use nlevels to verify it's a MultiIndex
    if new_index.nlevels > 1 and hasattr(new_index, "levels"):
        levels = list(new_index.levels)
    else:
        # Fallback: convert tuples to strings directly
        new_index = (map(str, entry) for entry in new_index)
        new_index = [sep.join([entry for entry in word if entry]) for word in new_index]
        setattr(df, axis, new_index)
        return df
    all_strings = all(map(is_string_dtype, levels))
    if all_strings:
        no_empty_string = all((entry != "").all() for entry in levels)
        if no_empty_string:
            new_index = new_index.map(sep.join)
            setattr(df, axis, new_index)
            return df
    new_index = (map(str, entry) for entry in new_index)
    new_index = [
        # faster to use a list comprehension within string.join
        # compared to a generator
        # https://stackoverflow.com/a/37782238
        sep.join([entry for entry in word if entry])
        for word in new_index
    ]
    setattr(df, axis, new_index)
    return df



"""Implementation of `get_one_to_one`."""

from __future__ import annotations

from typing import List

import numpy as np
import pandas as pd


def _get_one_to_one_value_order(series: pd.Series) -> np.ndarray:
    """Convert series values to ordered integer codes."""
    codes, _ = pd.factorize(series, sort=False)
    return codes


def get_one_to_one(df: pd.DataFrame) -> List[List[str]]:
    """Find column groups that map one-to-one with each other.

    Columns are grouped if they share identical value orderings, indicating
    a one-to-one mapping across rows.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "Lab_Test_Long": ["Cholesterol, LDL", "Cholesterol, LDL", "Glucose"],
        ...         "Lab_Test_Short": ["CLDL", "CLDL", "GLUC"],
        ...         "LOINC": [12345, 12345, 54321],
        ...         "Person": ["Sam", "Bill", "Sam"],
        ...     }
        ... )
        >>> janitor.get_one_to_one(df)
        [['Lab_Test_Long', 'Lab_Test_Short', 'LOINC']]

    Args:
        df: A pandas DataFrame.

    Raises:
        TypeError: If `df` is not a pandas DataFrame.
        ValueError: If `df` has no columns or duplicate column names.

    Returns:
        A list of column name groups that map one-to-one with each other.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df must be a pandas DataFrame.")
    if df.shape[1] == 0:
        raise ValueError("df must have at least one column.")
    if df.columns.duplicated().any():
        raise ValueError("df must have unique column names.")

    encoded = {col: _get_one_to_one_value_order(df[col]) for col in df.columns}
    remaining_cols = list(df.columns)
    groups: List[List[str]] = []

    while remaining_cols:
        nm1 = remaining_cols.pop(0)
        current_group = [nm1]
        for nm2 in remaining_cols[:]:
            if np.array_equal(encoded[nm1], encoded[nm2]):
                current_group.append(nm2)
                remaining_cols.remove(nm2)
        if len(current_group) > 1:
            groups.append(current_group)

    return groups



from __future__ import annotations

from typing import Any, Hashable

import pandas as pd
import pandas_flavor as pf

from janitor.utils import deprecated_alias, refactored_function


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `pd.DataFrame.astype` instead."
    )
)
@deprecated_alias(column="column_name")
def change_type(
    df: pd.DataFrame,
    column_name: Hashable | list[Hashable] | pd.Index,
    dtype: type,
    ignore_exception: bool = False,
) -> pd.DataFrame:
    """Change the type of a column.

    This method does not mutate the original DataFrame.

    Exceptions that are raised can be ignored. For example, if one has a mixed
    dtype column that has non-integer strings and integers, and you want to
    coerce everything to integers, you can optionally ignore the non-integer
    strings and replace them with `NaN` or keep the original value.

    Intended to be the method-chaining alternative to:

    ```python
    df[col] = df[col].astype(dtype)
    ```

    !!!note

        This function will be deprecated in a 1.x release.
        Please use `pd.DataFrame.astype` instead.

    Examples:
        Change the type of a column.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"col1": range(3), "col2": ["m", 5, True]})
        >>> df
           col1  col2
        0     0     m
        1     1     5
        2     2  True
        >>> df.change_type(
        ...     "col1",
        ...     dtype=str,
        ... ).change_type(
        ...     "col2",
        ...     dtype=float,
        ...     ignore_exception="fillna",
        ... )
          col1  col2
        0    0   NaN
        1    1   5.0
        2    2   1.0

        Change the type of multiple columns. To change the type of all columns,
        please use `DataFrame.astype` instead.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"col1": range(3), "col2": ["m", 5, True]})
        >>> df.change_type(["col1", "col2"], str)
          col1  col2
        0    0     m
        1    1     5
        2    2  True

    Args:
        df: A pandas DataFrame.
        column_name: The column(s) in the dataframe.
        dtype: The datatype to convert to. Should be one of the standard
            Python types, or a numpy datatype.
        ignore_exception: One of `{False, "fillna", "keep_values"}`.

    Raises:
        ValueError: If unknown option provided for `ignore_exception`.

    Returns:
        A pandas DataFrame with changed column types.
    """  # noqa: E501

    df = df.copy()  # avoid mutating the original DataFrame
    if not ignore_exception:
        df[column_name] = df[column_name].astype(dtype)
    elif ignore_exception == "keep_values":
        df[column_name] = df[column_name].astype(dtype, errors="ignore")
    elif ignore_exception == "fillna":
        if isinstance(column_name, Hashable):
            column_name = [column_name]
        df[column_name] = df[column_name].map(_convert, dtype=dtype)
    else:
        raise ValueError("Unknown option for ignore_exception")

    return df


def _convert(x: Any, dtype: type) -> Any:
    """Casts item `x` to `dtype` or None if not possible."""

    try:
        return dtype(x)
    except ValueError:
        return None



"""Implementation of the `row_to_names` function."""

from __future__ import annotations

from functools import singledispatch

import numpy as np
import pandas as pd
import pandas_flavor as pf

from janitor.utils import check, deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(row_number="row_numbers", remove_row="remove_rows")
def row_to_names(
    df: pd.DataFrame,
    row_numbers: int | list | slice = 0,
    remove_rows: bool = False,
    remove_rows_above: bool = False,
    reset_index: bool = False,
) -> pd.DataFrame:
    """Elevates a row, or rows, to be the column names of a DataFrame.

    This method does not mutate the original DataFrame.

    Contains options to remove the elevated row from the DataFrame along with
    removing the rows above the selected row.

    Examples:
        Replace column names with the first row and reset the index.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "a": ["nums", 6, 9],
        ...         "b": ["chars", "x", "y"],
        ...     }
        ... )
        >>> df
              a      b
        0  nums  chars
        1     6      x
        2     9      y
        >>> df.row_to_names(0, remove_rows=True, reset_index=True)
          nums chars
        0    6     x
        1    9     y
        >>> df.row_to_names([0, 1], remove_rows=True, reset_index=True)
          nums chars
             6     x
        0    9     y

        Remove rows above the elevated row and the elevated row itself.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "a": ["bla1", "nums", 6, 9],
        ...         "b": ["bla2", "chars", "x", "y"],
        ...     }
        ... )
        >>> df
              a      b
        0  bla1   bla2
        1  nums  chars
        2     6      x
        3     9      y
        >>> df.row_to_names(
        ...     1, remove_rows=True, remove_rows_above=True, reset_index=True
        ... )
          nums chars
        0    6     x
        1    9     y

    Args:
        df: A pandas DataFrame.
        row_numbers: Position of the row(s) containing the variable names.
            It can be an integer, a list or a slice.
            Defaults to 0 (first row).
        remove_rows: Whether the row(s) should be removed from the DataFrame.
        remove_rows_above: Whether the row(s) above the selected row should
            be removed from the DataFrame.
        reset_index: Whether the index should be reset on the returning DataFrame.

    Returns:
        A pandas DataFrame with set column names.
    """  # noqa: E501

    return _row_to_names(
        row_numbers,
        df=df,
        remove_rows=remove_rows,
        remove_rows_above=remove_rows_above,
        reset_index=reset_index,
    )


@singledispatch
def _row_to_names(
    row_numbers, df, remove_rows, remove_rows_above, reset_index
) -> pd.DataFrame:
    """
    Base function for row_to_names.
    """
    raise TypeError(
        "row_numbers should be either an integer, "
        "a slice or a list; "
        f"instead got type {type(row_numbers).__name__}"
    )


@_row_to_names.register(int)  # noqa: F811
def _row_to_names_dispatch(  # noqa: F811
    row_numbers, df, remove_rows, remove_rows_above, reset_index
):
    df_ = df[:]
    headers = df_.iloc[row_numbers]
    df_.columns = headers
    df_.columns.name = None
    if not remove_rows and not remove_rows_above and not reset_index:
        return df_
    if not remove_rows and not remove_rows_above and reset_index:
        return df_.reset_index(drop=True)

    len_df = len(df_)
    arrays = [arr._values for _, arr in df_.items()]
    if remove_rows_above and remove_rows:
        indexer = np.arange(row_numbers + 1, len_df)
    elif remove_rows_above:
        indexer = np.arange(row_numbers, len_df)
    elif remove_rows:
        indexer = np.arange(len_df)
        mask = np.ones(len_df, dtype=np.bool_)
        mask[row_numbers] = False
        indexer = indexer[mask]
    arrays = {num: arr[indexer] for num, arr in enumerate(arrays)}
    if reset_index:
        df_index = pd.RangeIndex(start=0, stop=indexer.size)
    else:
        df_index = df_.index[indexer]
    _df = pd.DataFrame(data=arrays, index=df_index, copy=False)
    _df.columns = df_.columns
    return _df


@_row_to_names.register(slice)  # noqa: F811
def _row_to_names_dispatch(  # noqa: F811
    row_numbers, df, remove_rows, remove_rows_above, reset_index
):
    if row_numbers.step is not None:
        raise ValueError(
            "The step argument for slice is not supported in row_to_names."
        )
    df_ = df[:]
    headers = df_.iloc[row_numbers]
    if isinstance(headers, pd.DataFrame) and (len(headers) == 1):
        headers = headers.squeeze()
        df_.columns = headers
        df_.columns.name = None
    else:
        headers = [array._values for _, array in headers.items()]
        headers = pd.MultiIndex.from_tuples(headers)
        df_.columns = headers
    if not remove_rows and not remove_rows_above and not reset_index:
        return df_
    if not remove_rows and not remove_rows_above and reset_index:
        return df_.reset_index(drop=True)
    len_df = len(df_)
    arrays = [arr._values for _, arr in df_.items()]
    if remove_rows_above and remove_rows:
        indexer = np.arange(row_numbers.stop, len_df)
    elif remove_rows_above:
        indexer = np.arange(row_numbers.start, len_df)
    elif remove_rows:
        indexer = np.arange(len_df)
        mask = np.ones(len_df, dtype=np.bool_)
        mask[row_numbers] = False
        indexer = indexer[mask]
    arrays = {num: arr[indexer] for num, arr in enumerate(arrays)}
    if reset_index:
        df_index = pd.RangeIndex(start=0, stop=indexer.size)
    else:
        df_index = df_.index[indexer]
    _df = pd.DataFrame(data=arrays, index=df_index, copy=False)
    _df.columns = df_.columns
    return _df


@_row_to_names.register(list)  # noqa: F811
def _row_to_names_dispatch(  # noqa: F811
    row_numbers, df, remove_rows, remove_rows_above, reset_index
):
    if remove_rows_above:
        raise ValueError(
            "The remove_rows_above argument is applicable "
            "only if the row_numbers argument is an integer "
            "or a slice."
        )

    for entry in row_numbers:
        check("entry in the row_numbers argument", entry, [int])

    df_ = df[:]
    headers = df_.iloc[row_numbers]
    if isinstance(headers, pd.DataFrame) and (len(headers) == 1):
        headers = headers.squeeze()
        df_.columns = headers
        df_.columns.name = None
    else:
        headers = [array._values for _, array in headers.items()]
        headers = pd.MultiIndex.from_tuples(headers)
        df_.columns = headers

    if not remove_rows and reset_index:
        return df_.reset_index(drop=True)
    if not remove_rows and not reset_index:
        return df_

    len_df = len(df_)
    arrays = [arr._values for _, arr in df_.items()]
    indexer = np.arange(len_df)
    mask = np.ones(len_df, dtype=np.bool_)
    mask[row_numbers] = False
    indexer = indexer[mask]

    arrays = {num: arr[indexer] for num, arr in enumerate(arrays)}
    if reset_index:
        df_index = pd.RangeIndex(start=0, stop=indexer.size)
    else:
        df_index = df_.index[indexer]
    _df = pd.DataFrame(data=arrays, index=df_index, copy=False)
    _df.columns = df_.columns
    return _df



"""
# General Functions

pyjanitor's general-purpose data cleaning functions.
"""

# NOTE: Instructions for future contributors:

# 1. Place the source code of the functions in a file named after the function.
# 2. Place utility functions in the same file.
# 3. If you use a utility function from another source file,
# please refactor it out to `janitor.functions.utils`.
# 4. Import the function into this file so that it shows up in the top-level API.
# 5. Sort the imports in alphabetical order.
# 6. Try to group related functions together (e.g. see `convert_date.py`)
# 7. Never import utils.

from .add_columns import add_columns
from .adorn import (
    adorn_ns,
    adorn_pct_formatting,
    adorn_percentages,
    adorn_rounding,
    adorn_title,
    adorn_totals,
)
from .alias import alias
from .also import also
from .bin_numeric import bin_numeric
from .case_when import case_when
from .change_index_dtype import change_index_dtype
from .change_type import change_type
from .statistical_tests import chisq_test, fisher_test
from .clean_names import clean_names
from .coalesce import coalesce
from .collapse_levels import collapse_levels
from .compare_df_cols import compare_df_cols, compare_df_cols_same, describe_class
from .complete import complete
from .concatenate_columns import concatenate_columns
from .conditional_join import conditional_join, get_join_indices, join_agg
from .convert_date import (
    convert_excel_date,
    convert_matlab_date,
    convert_to_date,
    convert_to_datetime,
    convert_unix_date,
    excel_time_to_numeric,
    sas_numeric_to_date,
)
from .count_cumulative_unique import count_cumulative_unique
from .currency_column_to_numeric import currency_column_to_numeric
from .deconcatenate_column import deconcatenate_column
from .drop_constant_columns import drop_constant_columns
from .drop_duplicate_columns import drop_duplicate_columns
from .dropnotnull import dropnotnull
from .encode_categorical import encode_categorical
from .expand_column import expand_column
from .expand_grid import cartesian_product, expand, expand_grid
from .explode_index import explode_index
from .factorize_columns import factorize_columns
from .fill import fill_direction, fill_empty
from .filter import filter_column_isin, filter_date, filter_on, filter_string
from .find_replace import find_replace
from .flag_nulls import flag_nulls
from .get_dupes import get_dupes
from .get_one_to_one import get_one_to_one
from .groupby_agg import groupby_agg
from .groupby_topk import groupby_topk
from .impute import impute
from .jitter import jitter
from .join_apply import join_apply
from .label_encode import label_encode
from .limit_column_characters import limit_column_characters
from .min_max_scale import min_max_scale
from .move import move
from .mutate import mutate
from .paste_skip_na import paste_skip_na
from .pivot import (
    pivot_longer,
    pivot_longer_spec,
    pivot_wider,
    pivot_wider_spec,
)
from .process_text import process_text
from .remove_columns import remove_columns
from .remove_empty import remove_empty
from .rename_columns import rename_column, rename_columns
from .reorder_columns import reorder_columns
from .rle_id import rle_id
from .round_half_up import round_half_up, signif_half_up
from .round_to_fraction import round_to_fraction
from .row_to_names import row_to_names
from .select import (
    DropLabel,
    get_columns,
    get_index_labels,
    select,
    select_columns,
    select_rows,
)
from .shuffle import shuffle
from .single_value import single_value
from .sort_column_value_order import sort_column_value_order
from .sort_naturally import sort_naturally
from .summarise import summarise
from .tabyl import tabyl
from .take_first import take_first
from .then import then
from .to_datetime import to_datetime
from .toset import toset
from .top_levels import top_levels
from .transform_columns import transform_column, transform_columns
from .truncate_datetime import truncate_datetime_dataframe
from .update_where import update_where
from .utils import (
    unionize_dataframe_categories,
)

__all__ = [
    "add_columns",
    "adorn_ns",
    "adorn_pct_formatting",
    "adorn_percentages",
    "adorn_rounding",
    "adorn_title",
    "adorn_totals",
    "alias",
    "also",
    "bin_numeric",
    "cartesian_product",
    "case_when",
    "change_type",
    "change_index_dtype",
    "chisq_test",
    "clean_names",
    "coalesce",
    "collapse_levels",
    "compare_df_cols",
    "compare_df_cols_same",
    "complete",
    "concatenate_columns",
    "conditional_join",
    "convert_excel_date",
    "convert_matlab_date",
    "convert_to_date",
    "convert_to_datetime",
    "convert_unix_date",
    "excel_time_to_numeric",
    "sas_numeric_to_date",
    "count_cumulative_unique",
    "currency_column_to_numeric",
    "deconcatenate_column",
    "describe_class",
    "drop_constant_columns",
    "drop_duplicate_columns",
    "dropnotnull",
    "encode_categorical",
    "expand",
    "expand_column",
    "expand_grid",
    "explode_index",
    "factorize_columns",
    "fisher_test",
    "fill_direction",
    "fill_empty",
    "filter_date",
    "filter_column_isin",
    "filter_on",
    "filter_string",
    "find_replace",
    "flag_nulls",
    "get_dupes",
    "get_one_to_one",
    "get_join_indices",
    "groupby_agg",
    "groupby_topk",
    "impute",
    "jitter",
    "join_apply",
    "join_agg",
    "label_encode",
    "limit_column_characters",
    "min_max_scale",
    "move",
    "mutate",
    "paste_skip_na",
    "pivot_longer",
    "pivot_longer_spec",
    "pivot_wider",
    "pivot_wider_spec",
    "process_text",
    "remove_columns",
    "remove_empty",
    "rename_column",
    "rename_columns",
    "reorder_columns",
    "rle_id",
    "round_half_up",
    "round_to_fraction",
    "row_to_names",
    "select_columns",
    "select_rows",
    "select",
    "shuffle",
    "signif_half_up",
    "single_value",
    "sort_column_value_order",
    "sort_naturally",
    "summarise",
    "tabyl",
    "take_first",
    "then",
    "to_datetime",
    "toset",
    "top_levels",
    "transform_column",
    "transform_columns",
    "truncate_datetime_dataframe",
    "update_where",
    "unionize_dataframe_categories",
    "DropLabel",
    "get_index_labels",
    "get_columns",
]



"""complete implementation for polars."""

from __future__ import annotations

from typing import Any

from janitor.utils import check, import_message

from .polars_flavor import register_dataframe_method, register_lazyframe_method

try:
    import polars as pl
    import polars.selectors as cs
    from polars._typing import ColumnNameOrSelector
except ImportError:
    import_message(
        submodule="polars",
        package="polars",
        conda_channel="conda-forge",
        pip_install=True,
    )


@register_lazyframe_method
@register_dataframe_method
def expand(
    df: pl.DataFrame | pl.LazyFrame,
    *columns: ColumnNameOrSelector,
    sort: bool = False,
    by: ColumnNameOrSelector = None,
) -> pl.DataFrame | pl.LazyFrame:
    """
    Creates a DataFrame from a cartesian combination of all inputs.

    Inspiration is from tidyr's expand() function.

    expand() is often useful with
    [pl.DataFrame.join](https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.join.html)
    to convert implicit
    missing values to explicit missing values - similar to
    [`complete`][janitor.polars.complete.complete].

    It can also be used to figure out which combinations are missing
    (e.g identify gaps in your DataFrame).

    The variable `columns` parameter can be a string,
    a ColumnSelector, a polars expression, or a polars Series.

    `expand` can also be applied to a LazyFrame.

    Examples:
        >>> import polars as pl
        >>> import janitor.polars
        >>> data = [
        ...     {"type": "apple", "year": 2010, "size": "XS"},
        ...     {"type": "orange", "year": 2010, "size": "S"},
        ...     {"type": "apple", "year": 2012, "size": "M"},
        ...     {"type": "orange", "year": 2010, "size": "S"},
        ...     {"type": "orange", "year": 2011, "size": "S"},
        ...     {"type": "orange", "year": 2012, "size": "M"},
        ... ]
        >>> df = pl.DataFrame(data)
        >>> df
        shape: (6, 3)
        ┌────────┬──────┬──────┐
        │ type   ┆ year ┆ size │
        │ ---    ┆ ---  ┆ ---  │
        │ str    ┆ i64  ┆ str  │
        ╞════════╪══════╪══════╡
        │ apple  ┆ 2010 ┆ XS   │
        │ orange ┆ 2010 ┆ S    │
        │ apple  ┆ 2012 ┆ M    │
        │ orange ┆ 2010 ┆ S    │
        │ orange ┆ 2011 ┆ S    │
        │ orange ┆ 2012 ┆ M    │
        └────────┴──────┴──────┘

        Get unique observations:
        >>> df.expand("type", sort=True)
        shape: (2, 1)
        ┌────────┐
        │ type   │
        │ ---    │
        │ str    │
        ╞════════╡
        │ apple  │
        │ orange │
        └────────┘
        >>> df.expand("size", sort=True)
        shape: (3, 1)
        ┌──────┐
        │ size │
        │ ---  │
        │ str  │
        ╞══════╡
        │ M    │
        │ S    │
        │ XS   │
        └──────┘
        >>> df.expand("type", "size", sort=True)
        shape: (6, 2)
        ┌────────┬──────┐
        │ type   ┆ size │
        │ ---    ┆ ---  │
        │ str    ┆ str  │
        ╞════════╪══════╡
        │ apple  ┆ M    │
        │ apple  ┆ S    │
        │ apple  ┆ XS   │
        │ orange ┆ M    │
        │ orange ┆ S    │
        │ orange ┆ XS   │
        └────────┴──────┘
        >>> with pl.Config(tbl_rows=-1):
        ...     df.expand("type", "size", "year", sort=True)
        shape: (18, 3)
        ┌────────┬──────┬──────┐
        │ type   ┆ size ┆ year │
        │ ---    ┆ ---  ┆ ---  │
        │ str    ┆ str  ┆ i64  │
        ╞════════╪══════╪══════╡
        │ apple  ┆ M    ┆ 2010 │
        │ apple  ┆ M    ┆ 2011 │
        │ apple  ┆ M    ┆ 2012 │
        │ apple  ┆ S    ┆ 2010 │
        │ apple  ┆ S    ┆ 2011 │
        │ apple  ┆ S    ┆ 2012 │
        │ apple  ┆ XS   ┆ 2010 │
        │ apple  ┆ XS   ┆ 2011 │
        │ apple  ┆ XS   ┆ 2012 │
        │ orange ┆ M    ┆ 2010 │
        │ orange ┆ M    ┆ 2011 │
        │ orange ┆ M    ┆ 2012 │
        │ orange ┆ S    ┆ 2010 │
        │ orange ┆ S    ┆ 2011 │
        │ orange ┆ S    ┆ 2012 │
        │ orange ┆ XS   ┆ 2010 │
        │ orange ┆ XS   ┆ 2011 │
        │ orange ┆ XS   ┆ 2012 │
        └────────┴──────┴──────┘

        Get observations that only occur in the data:
        >>> df.expand(pl.struct("type", "size"), sort=True).unnest("type")
        shape: (4, 2)
        ┌────────┬──────┐
        │ type   ┆ size │
        │ ---    ┆ ---  │
        │ str    ┆ str  │
        ╞════════╪══════╡
        │ apple  ┆ M    │
        │ apple  ┆ XS   │
        │ orange ┆ M    │
        │ orange ┆ S    │
        └────────┴──────┘
        >>> df.expand(pl.struct("type", "size", "year"), sort=True).unnest("type")
        shape: (5, 3)
        ┌────────┬──────┬──────┐
        │ type   ┆ size ┆ year │
        │ ---    ┆ ---  ┆ ---  │
        │ str    ┆ str  ┆ i64  │
        ╞════════╪══════╪══════╡
        │ apple  ┆ M    ┆ 2012 │
        │ apple  ┆ XS   ┆ 2010 │
        │ orange ┆ M    ┆ 2012 │
        │ orange ┆ S    ┆ 2010 │
        │ orange ┆ S    ┆ 2011 │
        └────────┴──────┴──────┘

        Expand the DataFrame to include new observations:
        >>> with pl.Config(tbl_rows=-1):
        ...     df.expand(
        ...         "type",
        ...         "size",
        ...         pl.int_range(2010, 2014).alias("new_year"),
        ...         sort=True,
        ...     )
        shape: (24, 3)
        ┌────────┬──────┬──────────┐
        │ type   ┆ size ┆ new_year │
        │ ---    ┆ ---  ┆ ---      │
        │ str    ┆ str  ┆ i64      │
        ╞════════╪══════╪══════════╡
        │ apple  ┆ M    ┆ 2010     │
        │ apple  ┆ M    ┆ 2011     │
        │ apple  ┆ M    ┆ 2012     │
        │ apple  ┆ M    ┆ 2013     │
        │ apple  ┆ S    ┆ 2010     │
        │ apple  ┆ S    ┆ 2011     │
        │ apple  ┆ S    ┆ 2012     │
        │ apple  ┆ S    ┆ 2013     │
        │ apple  ┆ XS   ┆ 2010     │
        │ apple  ┆ XS   ┆ 2011     │
        │ apple  ┆ XS   ┆ 2012     │
        │ apple  ┆ XS   ┆ 2013     │
        │ orange ┆ M    ┆ 2010     │
        │ orange ┆ M    ┆ 2011     │
        │ orange ┆ M    ┆ 2012     │
        │ orange ┆ M    ┆ 2013     │
        │ orange ┆ S    ┆ 2010     │
        │ orange ┆ S    ┆ 2011     │
        │ orange ┆ S    ┆ 2012     │
        │ orange ┆ S    ┆ 2013     │
        │ orange ┆ XS   ┆ 2010     │
        │ orange ┆ XS   ┆ 2011     │
        │ orange ┆ XS   ┆ 2012     │
        │ orange ┆ XS   ┆ 2013     │
        └────────┴──────┴──────────┘

        Filter for missing observations:
        >>> columns = ("type", "size", "year")
        >>> with pl.Config(tbl_rows=-1):
        ...     df.expand(*columns).join(df, how="anti", on=columns).sort(by=pl.all())
        shape: (13, 3)
        ┌────────┬──────┬──────┐
        │ type   ┆ size ┆ year │
        │ ---    ┆ ---  ┆ ---  │
        │ str    ┆ str  ┆ i64  │
        ╞════════╪══════╪══════╡
        │ apple  ┆ M    ┆ 2010 │
        │ apple  ┆ M    ┆ 2011 │
        │ apple  ┆ S    ┆ 2010 │
        │ apple  ┆ S    ┆ 2011 │
        │ apple  ┆ S    ┆ 2012 │
        │ apple  ┆ XS   ┆ 2011 │
        │ apple  ┆ XS   ┆ 2012 │
        │ orange ┆ M    ┆ 2010 │
        │ orange ┆ M    ┆ 2011 │
        │ orange ┆ S    ┆ 2012 │
        │ orange ┆ XS   ┆ 2010 │
        │ orange ┆ XS   ┆ 2011 │
        │ orange ┆ XS   ┆ 2012 │
        └────────┴──────┴──────┘

        Expand within each group, using `by`:
        >>> with pl.Config(tbl_rows=-1):
        ...     df.expand("year", "size", by="type", sort=True)
        shape: (10, 3)
        ┌────────┬──────┬──────┐
        │ type   ┆ year ┆ size │
        │ ---    ┆ ---  ┆ ---  │
        │ str    ┆ i64  ┆ str  │
        ╞════════╪══════╪══════╡
        │ apple  ┆ 2010 ┆ M    │
        │ apple  ┆ 2010 ┆ XS   │
        │ apple  ┆ 2012 ┆ M    │
        │ apple  ┆ 2012 ┆ XS   │
        │ orange ┆ 2010 ┆ M    │
        │ orange ┆ 2010 ┆ S    │
        │ orange ┆ 2011 ┆ M    │
        │ orange ┆ 2011 ┆ S    │
        │ orange ┆ 2012 ┆ M    │
        │ orange ┆ 2012 ┆ S    │
        └────────┴──────┴──────┘

    !!! info "New in version 0.28.0"

    Args:
        *columns: This refers to the columns to be completed.
            It can be a string or a column selector or a polars expression.
            A polars expression can be used to introduced new values,
            as long as the polars expression has a name that already exists
            in the DataFrame.
        sort: Sort the DataFrame based on *columns.
        by: Column(s) to group by.

    Returns:
        A polars DataFrame/LazyFrame.
    """  # noqa: E501
    if not columns:
        return df
    uniques, _ = _expand(df=df, columns=columns, by=by, sort=sort)
    return uniques


@register_lazyframe_method
@register_dataframe_method
def complete(
    df: pl.DataFrame | pl.LazyFrame,
    *columns: ColumnNameOrSelector,
    fill_value: dict | Any | pl.Expr = None,
    explicit: bool = True,
    sort: bool = False,
    by: ColumnNameOrSelector = None,
) -> pl.DataFrame | pl.LazyFrame:
    """
    Turns implicit missing values into explicit missing values

    It is modeled after tidyr's `complete` function.
    In a way, it is the inverse of `pl.drop_nulls`,
    as it exposes implicitly missing rows.

    If new values need to be introduced, a polars Expression
    or a polars Series with the new values can be passed,
    as long as the polars Expression/Series
    has a name that already exists in the DataFrame.

    `complete` can also be applied to a LazyFrame.

    Examples:
        >>> import polars as pl
        >>> import janitor.polars
        >>> df = pl.DataFrame(
        ...     dict(
        ...         group=(1, 2, 1, 2),
        ...         item_id=(1, 2, 2, 3),
        ...         item_name=("a", "a", "b", "b"),
        ...         value1=(1, None, 3, 4),
        ...         value2=range(4, 8),
        ...     )
        ... )
        >>> df
        shape: (4, 5)
        ┌───────┬─────────┬───────────┬────────┬────────┐
        │ group ┆ item_id ┆ item_name ┆ value1 ┆ value2 │
        │ ---   ┆ ---     ┆ ---       ┆ ---    ┆ ---    │
        │ i64   ┆ i64     ┆ str       ┆ i64    ┆ i64    │
        ╞═══════╪═════════╪═══════════╪════════╪════════╡
        │ 1     ┆ 1       ┆ a         ┆ 1      ┆ 4      │
        │ 2     ┆ 2       ┆ a         ┆ null   ┆ 5      │
        │ 1     ┆ 2       ┆ b         ┆ 3      ┆ 6      │
        │ 2     ┆ 3       ┆ b         ┆ 4      ┆ 7      │
        └───────┴─────────┴───────────┴────────┴────────┘

        Generate all possible combinations of
        `group`, `item_id`, and `item_name`
        (whether or not they appear in the data)
        >>> with pl.Config(tbl_rows=-1):
        ...     df.complete("group", "item_id", "item_name", sort=True)
        shape: (12, 5)
        ┌───────┬─────────┬───────────┬────────┬────────┐
        │ group ┆ item_id ┆ item_name ┆ value1 ┆ value2 │
        │ ---   ┆ ---     ┆ ---       ┆ ---    ┆ ---    │
        │ i64   ┆ i64     ┆ str       ┆ i64    ┆ i64    │
        ╞═══════╪═════════╪═══════════╪════════╪════════╡
        │ 1     ┆ 1       ┆ a         ┆ 1      ┆ 4      │
        │ 1     ┆ 1       ┆ b         ┆ null   ┆ null   │
        │ 1     ┆ 2       ┆ a         ┆ null   ┆ null   │
        │ 1     ┆ 2       ┆ b         ┆ 3      ┆ 6      │
        │ 1     ┆ 3       ┆ a         ┆ null   ┆ null   │
        │ 1     ┆ 3       ┆ b         ┆ null   ┆ null   │
        │ 2     ┆ 1       ┆ a         ┆ null   ┆ null   │
        │ 2     ┆ 1       ┆ b         ┆ null   ┆ null   │
        │ 2     ┆ 2       ┆ a         ┆ null   ┆ 5      │
        │ 2     ┆ 2       ┆ b         ┆ null   ┆ null   │
        │ 2     ┆ 3       ┆ a         ┆ null   ┆ null   │
        │ 2     ┆ 3       ┆ b         ┆ 4      ┆ 7      │
        └───────┴─────────┴───────────┴────────┴────────┘

        Cross all possible `group` values with the unique pairs of
        `(item_id, item_name)` that already exist in the data.
        >>> with pl.Config(tbl_rows=-1):
        ...     df.select(
        ...         "group", pl.struct("item_id", "item_name"), "value1", "value2"
        ...     ).complete("group", "item_id", sort=True).unnest("item_id")
        shape: (8, 5)
        ┌───────┬─────────┬───────────┬────────┬────────┐
        │ group ┆ item_id ┆ item_name ┆ value1 ┆ value2 │
        │ ---   ┆ ---     ┆ ---       ┆ ---    ┆ ---    │
        │ i64   ┆ i64     ┆ str       ┆ i64    ┆ i64    │
        ╞═══════╪═════════╪═══════════╪════════╪════════╡
        │ 1     ┆ 1       ┆ a         ┆ 1      ┆ 4      │
        │ 1     ┆ 2       ┆ a         ┆ null   ┆ null   │
        │ 1     ┆ 2       ┆ b         ┆ 3      ┆ 6      │
        │ 1     ┆ 3       ┆ b         ┆ null   ┆ null   │
        │ 2     ┆ 1       ┆ a         ┆ null   ┆ null   │
        │ 2     ┆ 2       ┆ a         ┆ null   ┆ 5      │
        │ 2     ┆ 2       ┆ b         ┆ null   ┆ null   │
        │ 2     ┆ 3       ┆ b         ┆ 4      ┆ 7      │
        └───────┴─────────┴───────────┴────────┴────────┘

        Fill in nulls:
        >>> with pl.Config(tbl_rows=-1):
        ...     df.select(
        ...         "group", pl.struct("item_id", "item_name"), "value1", "value2"
        ...     ).complete(
        ...         "group",
        ...         "item_id",
        ...         fill_value={"value1": 0, "value2": 99},
        ...         explicit=True,
        ...         sort=True,
        ...     ).unnest("item_id")
        shape: (8, 5)
        ┌───────┬─────────┬───────────┬────────┬────────┐
        │ group ┆ item_id ┆ item_name ┆ value1 ┆ value2 │
        │ ---   ┆ ---     ┆ ---       ┆ ---    ┆ ---    │
        │ i64   ┆ i64     ┆ str       ┆ i64    ┆ i64    │
        ╞═══════╪═════════╪═══════════╪════════╪════════╡
        │ 1     ┆ 1       ┆ a         ┆ 1      ┆ 4      │
        │ 1     ┆ 2       ┆ a         ┆ 0      ┆ 99     │
        │ 1     ┆ 2       ┆ b         ┆ 3      ┆ 6      │
        │ 1     ┆ 3       ┆ b         ┆ 0      ┆ 99     │
        │ 2     ┆ 1       ┆ a         ┆ 0      ┆ 99     │
        │ 2     ┆ 2       ┆ a         ┆ 0      ┆ 5      │
        │ 2     ┆ 2       ┆ b         ┆ 0      ┆ 99     │
        │ 2     ┆ 3       ┆ b         ┆ 4      ┆ 7      │
        └───────┴─────────┴───────────┴────────┴────────┘

        Limit the fill to only the newly created
        missing values with `explicit = FALSE`:
        >>> with pl.Config(tbl_rows=-1):
        ...     df.select(
        ...         "group", pl.struct("item_id", "item_name"), "value1", "value2"
        ...     ).complete(
        ...         "group",
        ...         "item_id",
        ...         fill_value={"value1": 0, "value2": 99},
        ...         explicit=False,
        ...         sort=True,
        ...     ).unnest("item_id").sort(pl.all())
        shape: (8, 5)
        ┌───────┬─────────┬───────────┬────────┬────────┐
        │ group ┆ item_id ┆ item_name ┆ value1 ┆ value2 │
        │ ---   ┆ ---     ┆ ---       ┆ ---    ┆ ---    │
        │ i64   ┆ i64     ┆ str       ┆ i64    ┆ i64    │
        ╞═══════╪═════════╪═══════════╪════════╪════════╡
        │ 1     ┆ 1       ┆ a         ┆ 1      ┆ 4      │
        │ 1     ┆ 2       ┆ a         ┆ 0      ┆ 99     │
        │ 1     ┆ 2       ┆ b         ┆ 3      ┆ 6      │
        │ 1     ┆ 3       ┆ b         ┆ 0      ┆ 99     │
        │ 2     ┆ 1       ┆ a         ┆ 0      ┆ 99     │
        │ 2     ┆ 2       ┆ a         ┆ null   ┆ 5      │
        │ 2     ┆ 2       ┆ b         ┆ 0      ┆ 99     │
        │ 2     ┆ 3       ┆ b         ┆ 4      ┆ 7      │
        └───────┴─────────┴───────────┴────────┴────────┘

        >>> df = pl.DataFrame(
        ...     {
        ...         "Year": [1999, 2000, 2004, 1999, 2004],
        ...         "Taxon": [
        ...             "Saccharina",
        ...             "Saccharina",
        ...             "Saccharina",
        ...             "Agarum",
        ...             "Agarum",
        ...         ],
        ...         "Abundance": [4, 5, 2, 1, 8],
        ...     }
        ... )
        >>> df
        shape: (5, 3)
        ┌──────┬────────────┬───────────┐
        │ Year ┆ Taxon      ┆ Abundance │
        │ ---  ┆ ---        ┆ ---       │
        │ i64  ┆ str        ┆ i64       │
        ╞══════╪════════════╪═══════════╡
        │ 1999 ┆ Saccharina ┆ 4         │
        │ 2000 ┆ Saccharina ┆ 5         │
        │ 2004 ┆ Saccharina ┆ 2         │
        │ 1999 ┆ Agarum     ┆ 1         │
        │ 2004 ┆ Agarum     ┆ 8         │
        └──────┴────────────┴───────────┘

        Expose missing years from 1999 to 2004 -
        pass a polars expression with the new dates,
        and ensure the expression's name already exists
        in the DataFrame:
        >>> expression = pl.int_range(1999, 2005).alias("Year")
        >>> with pl.Config(tbl_rows=-1):
        ...     df.complete(expression, "Taxon", sort=True)
        shape: (12, 3)
        ┌──────┬────────────┬───────────┐
        │ Year ┆ Taxon      ┆ Abundance │
        │ ---  ┆ ---        ┆ ---       │
        │ i64  ┆ str        ┆ i64       │
        ╞══════╪════════════╪═══════════╡
        │ 1999 ┆ Agarum     ┆ 1         │
        │ 1999 ┆ Saccharina ┆ 4         │
        │ 2000 ┆ Agarum     ┆ null      │
        │ 2000 ┆ Saccharina ┆ 5         │
        │ 2001 ┆ Agarum     ┆ null      │
        │ 2001 ┆ Saccharina ┆ null      │
        │ 2002 ┆ Agarum     ┆ null      │
        │ 2002 ┆ Saccharina ┆ null      │
        │ 2003 ┆ Agarum     ┆ null      │
        │ 2003 ┆ Saccharina ┆ null      │
        │ 2004 ┆ Agarum     ┆ 8         │
        │ 2004 ┆ Saccharina ┆ 2         │
        └──────┴────────────┴───────────┘

        Expose missing rows per group:
        >>> df = pl.DataFrame(
        ...     {
        ...         "state": ["CA", "CA", "HI", "HI", "HI", "NY", "NY"],
        ...         "year": [2010, 2013, 2010, 2012, 2016, 2009, 2013],
        ...         "value": [1, 3, 1, 2, 3, 2, 5],
        ...     }
        ... )
        >>> df
        shape: (7, 3)
        ┌───────┬──────┬───────┐
        │ state ┆ year ┆ value │
        │ ---   ┆ ---  ┆ ---   │
        │ str   ┆ i64  ┆ i64   │
        ╞═══════╪══════╪═══════╡
        │ CA    ┆ 2010 ┆ 1     │
        │ CA    ┆ 2013 ┆ 3     │
        │ HI    ┆ 2010 ┆ 1     │
        │ HI    ┆ 2012 ┆ 2     │
        │ HI    ┆ 2016 ┆ 3     │
        │ NY    ┆ 2009 ┆ 2     │
        │ NY    ┆ 2013 ┆ 5     │
        └───────┴──────┴───────┘
        >>> low = pl.col("year").min()
        >>> high = pl.col("year").max().add(1)
        >>> new_year_values = pl.int_range(low, high).alias("year")
        >>> with pl.Config(tbl_rows=-1):
        ...     df.complete(new_year_values, by="state", sort=True)
        shape: (16, 3)
        ┌───────┬──────┬───────┐
        │ state ┆ year ┆ value │
        │ ---   ┆ ---  ┆ ---   │
        │ str   ┆ i64  ┆ i64   │
        ╞═══════╪══════╪═══════╡
        │ CA    ┆ 2010 ┆ 1     │
        │ CA    ┆ 2011 ┆ null  │
        │ CA    ┆ 2012 ┆ null  │
        │ CA    ┆ 2013 ┆ 3     │
        │ HI    ┆ 2010 ┆ 1     │
        │ HI    ┆ 2011 ┆ null  │
        │ HI    ┆ 2012 ┆ 2     │
        │ HI    ┆ 2013 ┆ null  │
        │ HI    ┆ 2014 ┆ null  │
        │ HI    ┆ 2015 ┆ null  │
        │ HI    ┆ 2016 ┆ 3     │
        │ NY    ┆ 2009 ┆ 2     │
        │ NY    ┆ 2010 ┆ null  │
        │ NY    ┆ 2011 ┆ null  │
        │ NY    ┆ 2012 ┆ null  │
        │ NY    ┆ 2013 ┆ 5     │
        └───────┴──────┴───────┘


    !!! info "New in version 0.28.0"

    Args:
        *columns: This refers to the columns to be completed.
            It can be a string or a column selector or a polars expression.
            A polars expression can be used to introduced new values,
            as long as the polars expression has a name that already exists
            in the DataFrame.
        fill_value: Scalar value or polars expression to use instead of nulls
            for missing combinations. A dictionary, mapping columns names
            to a scalar value is also accepted.
        explicit: Determines if only implicitly missing values
            should be filled (`False`), or all nulls existing in the LazyFrame
            (`True`). `explicit` is applicable only
            if `fill_value` is not `None`.
        sort: Sort the DataFrame based on *columns.
        by: Column(s) to group by.
            The explicit missing rows are returned per group.

    Returns:
        A polars DataFrame/LazyFrame.
    """  # noqa: E501
    if not columns:
        return df
    return _complete(
        df=df,
        columns=columns,
        fill_value=fill_value,
        explicit=explicit,
        sort=sort,
        by=by,
    )


def _expand(
    df: pl.DataFrame | pl.LazyFrame,
    columns: tuple[ColumnNameOrSelector],
    sort: bool,
    by: ColumnNameOrSelector,
) -> pl.DataFrame | pl.LazyFrame:
    """
    This function computes the final output for the `complete` function.

    A DataFrame, with rows of missing values, if any, is returned.
    """

    check("sort", sort, [bool])
    _columns = []
    for column in columns:
        if isinstance(column, str):
            col = pl.col(column).unique()
        elif cs.is_selector(column):
            col = column.as_expr().unique()
        elif isinstance(column, pl.Expr):
            col = column.unique()
        elif isinstance(column, pl.Series):
            col = column.unique()
        else:
            raise TypeError(
                f"The argument passed to the columns parameter "
                "should either be a string, a column selector, "
                "a polars expression, or a polars Series; "
                f"instead got - {type(column)}."
            )
        if sort:
            col = col.sort()
        _columns.append(col)
    by_does_not_exist = by is None
    if by_does_not_exist:
        _columns = [column.implode() for column in _columns]
        uniques = df.select(_columns)
        uniques_schema = uniques.collect_schema()
        _columns = uniques_schema.names()
    else:
        uniques = df.group_by(by, maintain_order=sort).agg(_columns)
        uniques_schema = uniques.collect_schema()
        _columns = cs.expand_selector(uniques_schema, cs.exclude(by), strict=False)
    for column in _columns:
        uniques = uniques.explode(column)

    return uniques, uniques_schema


def _complete(
    df: pl.DataFrame | pl.LazyFrame,
    columns: tuple[ColumnNameOrSelector],
    fill_value: dict | Any | pl.Expr,
    explicit: bool,
    sort: bool,
    by: ColumnNameOrSelector,
) -> pl.DataFrame | pl.LazyFrame:
    """
    This function computes the final output for the `complete` function.

    A DataFrame, with rows of missing values, if any, is returned.
    """
    check("explicit", explicit, [bool])
    uniques, uniques_schema = _expand(df=df, columns=columns, sort=sort, by=by)

    df_columns = df.collect_schema()
    columns_to_fill = df_columns.keys() ^ uniques_schema.keys()
    if (fill_value is None) or not columns_to_fill:
        return uniques.join(df, on=uniques_schema.names(), how="left", coalesce=True)
    idx = None
    columns_to_select = df_columns.names()
    if not explicit:
        idx = "".join(columns_to_select)
        idx = f"{idx}_"
        df = df.with_row_index(name=idx)
    df = uniques.join(df, on=uniques_schema.names(), how="left", coalesce=True)
    # exclude columns that were not used
    # to generate the combinations
    exclude_columns = uniques_schema.names()
    if idx:
        exclude_columns.append(idx)
    _columns = set(columns_to_select).difference(exclude_columns)

    if isinstance(fill_value, dict):
        _columns = _columns.intersection(fill_value)
        fill_value = [
            pl.col(column_name).fill_null(value=fill_value[column_name])
            for column_name in _columns
        ]
    else:
        fill_value = [pl.col(column).fill_null(value=fill_value) for column in _columns]
    if not explicit:
        condition = pl.col(idx).is_null()
        fill_value = [
            pl.when(condition).then(_fill_value).otherwise(pl.col(column_name))
            for column_name, _fill_value in zip(_columns, fill_value)
        ]
    df = df.with_columns(fill_value)

    return df.select(columns_to_select)



"""Implementation for expand_column."""

from typing import Hashable

import pandas as pd
import pandas_flavor as pf

from janitor.utils import deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(column="column_name")
def expand_column(
    df: pd.DataFrame,
    column_name: Hashable,
    sep: str = "|",
    concat: bool = True,
) -> pd.DataFrame:
    """Expand a categorical column with multiple labels into dummy-coded columns.

    Super sugary syntax that wraps `pandas.Series.str.get_dummies`.

    This method does not mutate the original DataFrame.

    Examples:
        Functional usage syntax:

        >>> import pandas as pd
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": ["A, B", "B, C, D", "E, F", "A, E, F"],
        ...         "col2": [1, 2, 3, 4],
        ...     }
        ... )
        >>> df = expand_column(
        ...     df,
        ...     column_name="col1",
        ...     sep=", ",  # note space in sep
        ... )
        >>> df
              col1  col2  A  B  C  D  E  F
        0     A, B     1  1  1  0  0  0  0
        1  B, C, D     2  0  1  1  1  0  0
        2     E, F     3  0  0  0  0  1  1
        3  A, E, F     4  1  0  0  0  1  1

        Method chaining syntax:

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": ["A, B", "B, C, D", "E, F", "A, E, F"],
        ...         "col2": [1, 2, 3, 4],
        ...     }
        ... ).expand_column(column_name="col1", sep=", ")
        >>> df
              col1  col2  A  B  C  D  E  F
        0     A, B     1  1  1  0  0  0  0
        1  B, C, D     2  0  1  1  1  0  0
        2     E, F     3  0  0  0  0  1  1
        3  A, E, F     4  1  0  0  0  1  1

    Args:
        df: A pandas DataFrame.
        column_name: Which column to expand.
        sep: The delimiter, same to
            `pandas.Series.str.get_dummies`'s `sep`.
        concat: Whether to return the expanded column concatenated to
            the original dataframe (`concat=True`), or to return it standalone
            (`concat=False`).

    Returns:
        A pandas DataFrame with an expanded column.
    """  # noqa: E501
    expanded_df = df[column_name].str.get_dummies(sep=sep)
    if concat:
        return df.join(expanded_df)
    return expanded_df



"""Implementation source for `dropnotnull`."""

from typing import Hashable

import pandas as pd
import pandas_flavor as pf

from janitor.utils import deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(column="column_name")
def dropnotnull(df: pd.DataFrame, column_name: Hashable) -> pd.DataFrame:
    """Drop rows that do *not* have null values in the given column.

    This method does not mutate the original DataFrame.

    Examples:
        >>> import numpy as np
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [1.0, np.nan, 3.0], "b": [None, "y", "z"]})
        >>> df
             a     b
        0  1.0   NaN
        1  NaN     y
        2  3.0     z
        >>> df.dropnotnull("a")
            a  b
        1 NaN  y
        >>> df.dropnotnull("b")
             a     b
        0  1.0   NaN

    Args:
        df: A pandas DataFrame.
        column_name: The column name to drop rows from.

    Returns:
        A pandas DataFrame with dropped rows.
    """
    return df[pd.isna(df[column_name])]



"""Implementation of count_cumulative_unique."""

from typing import Hashable

import numpy as np
import pandas as pd
import pandas_flavor as pf

from janitor.utils import check_column


@pf.register_dataframe_method
def count_cumulative_unique(
    df: pd.DataFrame,
    column_name: Hashable,
    dest_column_name: str,
    case_sensitive: bool = True,
) -> pd.DataFrame:
    """Generates a running total of cumulative unique values in a given column.

    A new column will be created containing a running
    count of unique values in the specified column.
    If `case_sensitive` is `True`, then the case of
    any letters will matter (i.e., `a != A`);
    otherwise, the case of any letters will not matter.

    This method does not mutate the original DataFrame.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "letters": list("aabABb"),
        ...         "numbers": range(4, 10),
        ...     }
        ... )
        >>> df
          letters  numbers
        0       a        4
        1       a        5
        2       b        6
        3       A        7
        4       B        8
        5       b        9
        >>> df.count_cumulative_unique(
        ...     column_name="letters",
        ...     dest_column_name="letters_unique_count",
        ... )
          letters  numbers  letters_unique_count
        0       a        4                     1
        1       a        5                     1
        2       b        6                     2
        3       A        7                     3
        4       B        8                     4
        5       b        9                     4

        Cumulative counts, ignoring casing.

        >>> df.count_cumulative_unique(
        ...     column_name="letters",
        ...     dest_column_name="letters_unique_count",
        ...     case_sensitive=False,
        ... )
          letters  numbers  letters_unique_count
        0       a        4                     1
        1       a        5                     1
        2       b        6                     2
        3       A        7                     2
        4       B        8                     2
        5       b        9                     2

    Args:
        df: A pandas DataFrame.
        column_name: Name of the column containing values from which a
            running count of unique values will be created.
        dest_column_name: The name of the new column containing the
            cumulative count of unique values that will be created.
        case_sensitive: Whether or not uppercase and lowercase letters
            will be considered equal. Only valid with string-like columns.

    Raises:
        TypeError: If `case_sensitive` is False when counting a non-string
            `column_name`.

    Returns:
        A pandas DataFrame with a new column containing a cumulative
            count of unique values from another column.
    """
    check_column(df, column_name)
    check_column(df, dest_column_name, present=False)

    counter = df[column_name]
    if not case_sensitive:
        try:
            # Make it so that the the same uppercase and lowercase
            # letter are treated as one unique value
            counter = counter.str.lower()
        except (AttributeError, TypeError) as e:
            # AttributeError is raised by pandas when .str is used on
            # non-string types, e.g. int.
            # TypeError is raised by pandas when .str.lower is used on a
            # forbidden string type, e.g. bytes.
            raise TypeError(
                "case_sensitive=False can only be used with a string-like "
                f"type. Column {column_name} is {counter.dtype} type."
            ) from e

    counter = counter.groupby(counter, sort=False).cumcount().to_numpy(copy=False)
    counter = np.cumsum(counter == 0)

    return df.assign(**{dest_column_name: counter})



"""Implementation of the `toset` function."""

from typing import Set

import pandas as pd
import pandas_flavor as pf

from janitor.utils import refactored_function


@pf.register_series_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `set(df[column])` instead."
    )
)
def toset(series: pd.Series) -> Set:
    """Return a set of the values.

    !!!note

        This function will be deprecated in a 1.x release.
        Please use `set(df[column])` instead.

    These are each a scalar type, which is a Python scalar
    (for str, int, float) or a pandas scalar
    (for Timestamp/Timedelta/Interval/Period)

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> s = pd.Series([1, 2, 3, 5, 5], index=["a", "b", "c", "d", "e"])
        >>> s
        a    1
        b    2
        c    3
        d    5
        e    5
        dtype: int64
        >>> s.toset()
        {1, 2, 3, 5}

    Args:
        series: A pandas series.

    Returns:
        A set of values.
    """

    return set(series.tolist())



"""DataDescription class for the DataDescription accessor."""

from typing import Dict, List, Union

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_accessor("data_description")
class DataDescription:
    """High-level description of data present in this DataFrame.

    This is a custom data accessor.
    """

    # Key used to store descriptions in DataFrame.attrs, which persists
    # across accessor instantiations (pandas 3.0+ no longer caches
    # accessor instances).
    _ATTRS_KEY = "_data_description_desc"

    def __init__(self, data):
        self._data = data

    @property
    def _desc(self) -> dict:
        """Get descriptions from DataFrame.attrs for persistence."""
        return self._data.attrs.get(self._ATTRS_KEY, {})

    @_desc.setter
    def _desc(self, value: dict):
        """Store descriptions in DataFrame.attrs for persistence."""
        self._data.attrs[self._ATTRS_KEY] = value

    def _get_data_df(self) -> pd.DataFrame:
        """Get a table of descriptive information in a DataFrame format.

        :returns: A DataFrame containing the descriptive information.
        """
        df = self._data

        data_dict = {}
        data_dict["column_name"] = df.columns.tolist()
        data_dict["type"] = df.dtypes.tolist()
        data_dict["count"] = df.count().tolist()
        data_dict["pct_missing"] = (1 - (df.count() / len(df))).tolist()
        data_dict["description"] = [self._desc.get(c, "") for c in df.columns]

        return pd.DataFrame(data_dict).set_index("column_name")

    @property
    def df(self) -> pd.DataFrame:
        """Get a table of descriptive information in a DataFrame format."""
        return self._get_data_df()

    def __repr__(self):
        """Human-readable representation of the `DataDescription` object."""
        return str(self._get_data_df())

    def display(self):
        """Print the table of descriptive information about this DataFrame."""
        print(self)

    def set_description(self, desc: Union[List, Dict]):
        """Update the description for each of the columns in the DataFrame.

        Args:
            desc: The structure containing the descriptions to update

        Raises:
            ValueError: If length of description list does not match
                number of columns in DataFrame.
        """
        if isinstance(desc, list):
            if len(desc) != len(self._data.columns):
                raise ValueError(
                    "Length of description list "
                    f"({len(desc)}) does not match number of columns in "
                    f"DataFrame ({len(self._data.columns)})"
                )

            self._desc = dict(zip(self._data.columns, desc))

        elif isinstance(desc, dict):
            self._desc = desc



from typing import Callable, Dict, Union

import pandas as pd
import pandas_flavor as pf

from janitor.utils import check_column, deprecated_alias, refactored_function


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `pd.DataFrame.rename` instead."
    )
)
@deprecated_alias(old="old_column_name", new="new_column_name")
def rename_column(
    df: pd.DataFrame,
    old_column_name: str,
    new_column_name: str,
) -> pd.DataFrame:
    """Rename a column in place.

    This method does not mutate the original DataFrame.

    !!!note

        This function will be deprecated in a 1.x release.
        Please use `pd.DataFrame.rename` instead.

    This is just syntactic sugar/a convenience function for renaming one column at a time.
    If you are convinced that there are multiple columns in need of changing,
    then use the `pandas.DataFrame.rename` method.

    Examples:
        Change the name of column 'a' to 'a_new'.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": list(range(3)), "b": list("abc")})
        >>> df.rename_column(old_column_name="a", new_column_name="a_new")
           a_new  b
        0      0  a
        1      1  b
        2      2  c

    Args:
        df: The pandas DataFrame object.
        old_column_name: The old column name.
        new_column_name: The new column name.

    Returns:
        A pandas DataFrame with renamed columns.
    """  # noqa: E501

    check_column(df, [old_column_name])

    return df.rename(columns={old_column_name: new_column_name})


@pf.register_dataframe_method
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `pd.DataFrame.rename` instead."
    )
)
def rename_columns(
    df: pd.DataFrame,
    new_column_names: Union[Dict, None] = None,
    function: Callable = None,
) -> pd.DataFrame:
    """Rename columns.

    This method does not mutate the original DataFrame.

    !!!note

        This function will be deprecated in a 1.x release.
        Please use `pd.DataFrame.rename` instead.

    One of the `new_column_names` or `function` are a required parameter.
    If both are provided, then `new_column_names` takes priority and `function`
    is never executed.

    Examples:
        Rename columns using a dictionary which maps old names to new names.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": list(range(3)), "b": list("xyz")})
        >>> df
           a  b
        0  0  x
        1  1  y
        2  2  z
        >>> df.rename_columns(new_column_names={"a": "a_new", "b": "b_new"})
           a_new b_new
        0      0     x
        1      1     y
        2      2     z

        Rename columns using a generic callable.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": list(range(3)), "b": list("xyz")})
        >>> df.rename_columns(function=str.upper)
           A  B
        0  0  x
        1  1  y
        2  2  z

    Args:
        df: The pandas DataFrame object.
        new_column_names: A dictionary of old and new column names.
        function: A function which should be applied to all the columns.

    Raises:
        ValueError: If both `new_column_names` and `function` are None.

    Returns:
        A pandas DataFrame with renamed columns.
    """  # noqa: E501

    if new_column_names is None and function is None:
        raise ValueError("One of new_column_names or function must be provided")

    if new_column_names is not None:
        check_column(df, new_column_names)
        return df.rename(columns=new_column_names)

    return df.rename(mapper=function, axis="columns")



"""Implementation of the `toset` function."""

from __future__ import annotations

from typing import Any

import pandas as pd
import pandas_flavor as pf


@pf.register_series_method
def alias(series: pd.Series, alias: Any = None) -> pd.Series:
    """Return a Series with a new name. Accepts either a scalar or a callable.


    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> s = pd.Series([1, 2, 3], name="series")
        >>> s
        0    1
        1    2
        2    3
        Name: series, dtype: int64
        >>> s.alias("series_new")
        0    1
        1    2
        2    3
        Name: series_new, dtype: int64
        >>> s.alias(str.upper)
        0    1
        1    2
        2    3
        Name: SERIES, dtype: int64

    Args:
        series: A pandas Series.
        alias: scalar or callable to create a new name for the pandas Series.

    Returns:
        A new pandas Series.
    """
    series = series[:]
    if alias is None:
        return series
    if callable(alias):
        alias = alias(series.name)
    series.name = alias
    return series



"""Implementation of take_first function."""

from typing import Hashable, Iterable, Union

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_method
def take_first(
    df: pd.DataFrame,
    subset: Union[Hashable, Iterable[Hashable]],
    by: Hashable,
    ascending: bool = True,
) -> pd.DataFrame:
    """Take the first row within each group specified by `subset`.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": ["x", "x", "y", "y"], "b": [0, 1, 2, 3]})
        >>> df
           a  b
        0  x  0
        1  x  1
        2  y  2
        3  y  3
        >>> df.take_first(subset="a", by="b")
           a  b
        0  x  0
        2  y  2

    Args:
        df: A pandas DataFrame.
        subset: Column(s) defining the group.
        by: Column to sort by.
        ascending: Whether or not to sort in ascending order, `bool`.

    Returns:
        A pandas DataFrame.
    """
    result = df.sort_values(by=by, ascending=ascending).drop_duplicates(
        subset=subset, keep="first"
    )

    return result
